# Copyright (C) 2015 Canonical Ltd.
# Copyright (C) 2006-2024 Broadcom. All Rights Reserved.
# Broadcom Confidential. The term "Broadcom" refers to Broadcom Inc.
# and/or its subsidiaries.
#
# Author: Sankar Tanguturi <stanguturi@vmware.com>
#         Pengpeng Sun <pengpeng.sun@broadcom.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

import logging
import os
import sys
import tempfile
import textwrap

from cloudinit.sources.helpers.vmware.imc.boot_proto import BootProtoEnum
from cloudinit.sources.helpers.vmware.imc.config import Config
from cloudinit.sources.helpers.vmware.imc.config_file import (
    ConfigFile as WrappedConfigFile,
)
from cloudinit.sources.helpers.vmware.imc.config_nic import NicConfigurator
from cloudinit.sources.helpers.vmware.imc.guestcust_util import (
    get_network_data_from_vmware_cust_cfg,
    get_non_network_data_from_vmware_cust_cfg,
)
from tests.helpers import cloud_init_project_dir
from tests.unittests.helpers import CiTestCase

logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)
logger = logging.getLogger(__name__)


def ConfigFile(path: str):
    return WrappedConfigFile(cloud_init_project_dir(path))


class TestVmwareConfigFile(CiTestCase):
    def test_utility_methods(self):
        """Tests basic utility methods of ConfigFile class"""
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")

        cf.clear()

        self.assertEqual(0, len(cf), "clear size")

        cf._insertKey("  PASSWORD|-PASS ", "  foo  ")
        cf._insertKey("BAR", "   ")

        self.assertEqual(2, len(cf), "insert size")
        self.assertEqual("foo", cf["PASSWORD|-PASS"], "password")
        self.assertTrue("PASSWORD|-PASS" in cf, "hasPassword")
        self.assertFalse("FOO" in cf, "hasFoo")
        self.assertTrue("BAR" in cf, "hasBar")

    def test_configfile_without_instance_id(self):
        """
        Tests instance id is None when configuration file has no instance id
        """
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")
        conf = Config(cf)

        (md1, _) = get_non_network_data_from_vmware_cust_cfg(conf)
        self.assertFalse("instance-id" in md1)

        (md2, _) = get_non_network_data_from_vmware_cust_cfg(conf)
        self.assertFalse("instance-id" in md2)

    def test_configfile_with_instance_id(self):
        """Tests instance id get from configuration file"""
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic-instance-id.cfg")
        conf = Config(cf)

        (md1, _) = get_non_network_data_from_vmware_cust_cfg(conf)
        self.assertEqual(md1["instance-id"], conf.instance_id, "instance-id")

        (md2, _) = get_non_network_data_from_vmware_cust_cfg(conf)
        self.assertEqual(md2["instance-id"], conf.instance_id, "instance-id")

    def test_configfile_static_2nics(self):
        """Tests Config class for a configuration with two static NICs."""
        cf = ConfigFile("tests/data/vmware/cust-static-2nic.cfg")

        conf = Config(cf)

        self.assertEqual("myhost1", conf.host_name, "hostName")
        self.assertEqual("Africa/Abidjan", conf.timezone, "tz")

        self.assertEqual(
            ["10.20.145.1", "10.20.145.2"], conf.name_servers, "dns"
        )
        self.assertEqual(
            ["eng.vmware.com", "proxy.vmware.com"],
            conf.dns_suffixes,
            "suffixes",
        )

        nics = conf.nics
        ipv40 = nics[0].staticIpv4

        self.assertEqual(2, len(nics), "nics")
        self.assertEqual("NIC1", nics[0].name, "nic0")
        self.assertEqual("00:50:56:a6:8c:08", nics[0].mac, "mac0")
        self.assertEqual(BootProtoEnum.STATIC, nics[0].bootProto, "bootproto0")
        self.assertEqual("10.20.87.154", ipv40[0].ip, "ipv4Addr0")
        self.assertEqual("255.255.252.0", ipv40[0].netmask, "ipv4Mask0")
        self.assertEqual(2, len(ipv40[0].gateways), "ipv4Gw0")
        self.assertEqual("10.20.87.253", ipv40[0].gateways[0], "ipv4Gw0_0")
        self.assertEqual("10.20.87.105", ipv40[0].gateways[1], "ipv4Gw0_1")

        self.assertEqual(1, len(nics[0].staticIpv6), "ipv6Cnt0")
        self.assertEqual(
            "fc00:10:20:87::154", nics[0].staticIpv6[0].ip, "ipv6Addr0"
        )

        self.assertEqual("NIC2", nics[1].name, "nic1")
        self.assertTrue(not nics[1].staticIpv6, "ipv61 dhcp")

    def test_config_file_dhcp_2nics(self):
        """Tests Config class for a configuration with two DHCP NICs."""
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")

        conf = Config(cf)
        nics = conf.nics
        self.assertEqual(2, len(nics), "nics")
        self.assertEqual("NIC1", nics[0].name, "nic0")
        self.assertEqual("00:50:56:a6:8c:08", nics[0].mac, "mac0")
        self.assertEqual(BootProtoEnum.DHCP, nics[0].bootProto, "bootproto0")

    def test_config_password(self):
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")

        cf._insertKey("PASSWORD|-PASS", "test-password")
        cf._insertKey("PASSWORD|RESET", "no")

        conf = Config(cf)
        self.assertEqual("test-password", conf.admin_password, "password")
        self.assertFalse(conf.reset_password, "do not reset password")

    def test_config_reset_passwd(self):
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")

        cf._insertKey("PASSWORD|-PASS", "test-password")
        cf._insertKey("PASSWORD|RESET", "random")

        conf = Config(cf)
        with self.assertRaises(ValueError):
            pw = conf.reset_password
            self.assertIsNone(pw)

        cf.clear()
        cf._insertKey("PASSWORD|RESET", "yes")
        self.assertEqual(1, len(cf), "insert size")

        conf = Config(cf)
        self.assertTrue(conf.reset_password, "reset password")

    def test_get_config_nameservers(self):
        """Tests DNS and nameserver settings in a configuration."""
        cf = ConfigFile("tests/data/vmware/cust-static-2nic.cfg")

        config = Config(cf)

        network_config = get_network_data_from_vmware_cust_cfg(config, False)

        self.assertEqual(2, network_config.get("version"))

        ethernets = network_config.get("ethernets")

        for _, config in ethernets.items():
            self.assertTrue(isinstance(config, dict))
            name_servers = config.get("nameservers").get("addresses")
            dns_suffixes = config.get("nameservers").get("search")
            self.assertEqual(
                ["10.20.145.1", "10.20.145.2"], name_servers, "dns"
            )
            self.assertEqual(
                ["eng.vmware.com", "proxy.vmware.com"],
                dns_suffixes,
                "suffixes",
            )

    def test_get_config_dns_suffixes(self):
        """Tests if get_network_from_vmware_cust_cfg properly
        generates nameservers and dns settings from a
        specified configuration"""
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")

        config = Config(cf)

        network_config = get_network_data_from_vmware_cust_cfg(config, False)

        self.assertEqual(2, network_config.get("version"))

        ethernets = network_config.get("ethernets")

        for _, config in ethernets.items():
            self.assertTrue(isinstance(config, dict))
            name_servers = config.get("nameservers").get("addresses")
            dns_suffixes = config.get("nameservers").get("search")
            self.assertEqual(None, name_servers, "dns")
            self.assertEqual(["eng.vmware.com"], dns_suffixes, "suffixes")

    def test_get_nics_list_dhcp(self):
        """Tests if NicConfigurator properly calculates ethernets
        for a configuration with a list of DHCP NICs"""
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")

        config = Config(cf)

        nicConfigurator = NicConfigurator(
            config.nics, config.name_servers, config.dns_suffixes, False
        )
        ethernets_dict = nicConfigurator.generate()

        self.assertTrue(isinstance(ethernets_dict, dict))
        self.assertEqual(2, len(ethernets_dict), "number of ethernets")

        for name, config in ethernets_dict.items():
            if name == "NIC1":
                self.assertEqual(
                    "00:50:56:a6:8c:08",
                    config.get("match").get("macaddress"),
                    "mac address of NIC1",
                )
                self.assertEqual(
                    True, config.get("wakeonlan"), "wakeonlan of NIC1"
                )
                self.assertEqual(
                    True, config.get("dhcp4"), "DHCPv4 enablement of NIC1"
                )
                self.assertEqual(
                    False,
                    config.get("dhcp4-overrides").get("use-dns"),
                    "use-dns enablement for dhcp4-overrides of NIC1",
                )
            if name == "NIC2":
                self.assertEqual(
                    "00:50:56:a6:5a:de",
                    config.get("match").get("macaddress"),
                    "mac address of NIC2",
                )
                self.assertEqual(
                    True, config.get("wakeonlan"), "wakeonlan of NIC2"
                )
                self.assertEqual(
                    True, config.get("dhcp4"), "DHCPv4 enablement of NIC2"
                )
                self.assertEqual(
                    False,
                    config.get("dhcp4-overrides").get("use-dns"),
                    "use-dns enablement for dhcp4-overrides of NIC2",
                )

    def test_get_nics_list_static(self):
        """Tests if NicConfigurator properly calculates ethernets
        for a configuration with 2 static NICs"""
        cf = ConfigFile("tests/data/vmware/cust-static-2nic.cfg")

        config = Config(cf)

        nicConfigurator = NicConfigurator(
            config.nics, config.name_servers, config.dns_suffixes, False
        )
        ethernets_dict = nicConfigurator.generate()

        self.assertTrue(isinstance(ethernets_dict, dict))
        self.assertEqual(2, len(ethernets_dict), "number of ethernets")

        for name, config in ethernets_dict.items():
            print(config)
            if name == "NIC1":
                self.assertEqual(
                    "00:50:56:a6:8c:08",
                    config.get("match").get("macaddress"),
                    "mac address of NIC1",
                )
                self.assertEqual(
                    True, config.get("wakeonlan"), "wakeonlan of NIC1"
                )
                self.assertEqual(
                    False, config.get("dhcp4"), "DHCPv4 enablement of NIC1"
                )
                self.assertEqual(
                    False, config.get("dhcp6"), "DHCPv6 enablement of NIC1"
                )
                self.assertEqual(
                    ["10.20.87.154/22", "fc00:10:20:87::154/64"],
                    config.get("addresses"),
                    "IP addresses of NIC1",
                )
                self.assertEqual(
                    [
                        {"to": "10.20.84.0/22", "via": "10.20.87.253"},
                        {"to": "10.20.84.0/22", "via": "10.20.87.105"},
                        {
                            "to": "fc00:10:20:87::/64",
                            "via": "fc00:10:20:87::253",
                        },
                    ],
                    config.get("routes"),
                    "routes of NIC1",
                )
            if name == "NIC2":
                self.assertEqual(
                    "00:50:56:a6:ef:7d",
                    config.get("match").get("macaddress"),
                    "mac address of NIC2",
                )
                self.assertEqual(
                    True, config.get("wakeonlan"), "wakeonlan of NIC2"
                )
                self.assertEqual(
                    False, config.get("dhcp4"), "DHCPv4 enablement of NIC2"
                )
                self.assertEqual(
                    ["192.168.6.102/16"],
                    config.get("addresses"),
                    "IP addresses of NIC2",
                )
                self.assertEqual(
                    [
                        {"to": "192.168.0.0/16", "via": "192.168.0.10"},
                    ],
                    config.get("routes"),
                    "routes of NIC2",
                )

    def test_custom_script(self):
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")
        conf = Config(cf)
        self.assertIsNone(conf.custom_script_name)
        cf._insertKey("CUSTOM-SCRIPT|SCRIPT-NAME", "test-script")
        conf = Config(cf)
        self.assertEqual("test-script", conf.custom_script_name)

    def test_post_gc_status(self):
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")
        conf = Config(cf)
        self.assertFalse(conf.post_gc_status)
        cf._insertKey("MISC|POST-GC-STATUS", "YES")
        conf = Config(cf)
        self.assertTrue(conf.post_gc_status)

    def test_no_default_run_post_script(self):
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")
        conf = Config(cf)
        self.assertFalse(conf.default_run_post_script)
        cf._insertKey("MISC|DEFAULT-RUN-POST-CUST-SCRIPT", "NO")
        conf = Config(cf)
        self.assertFalse(conf.default_run_post_script)

    def test_yes_default_run_post_script(self):
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")
        cf._insertKey("MISC|DEFAULT-RUN-POST-CUST-SCRIPT", "yes")
        conf = Config(cf)
        self.assertTrue(conf.default_run_post_script)


class TestVmwareNetConfig(CiTestCase):
    """Test conversion of vmware config to cloud-init config."""

    maxDiff = None

    def _get_NicConfigurator(self, text):
        fp = None
        try:
            with tempfile.NamedTemporaryFile(
                mode="w", dir=self.tmp_dir(), delete=False
            ) as fp:
                fp.write(text)
                fp.close()
            cfg = Config(ConfigFile(fp.name))
            return NicConfigurator(
                cfg.nics,
                cfg.name_servers,
                cfg.dns_suffixes,
                use_system_devices=False,
            )
        finally:
            if fp:
                os.unlink(fp.name)

    def test_static_nic_without_ipv4_netmask(self):
        """netmask is optional for static ipv4 configuration."""
        config = textwrap.dedent(
            """\
            [NETWORK]
            NETWORKING = yes
            BOOTPROTO = dhcp
            HOSTNAME = myhost1
            DOMAINNAME = eng.vmware.com

            [NIC-CONFIG]
            NICS = NIC1

            [NIC1]
            MACADDR = 00:50:56:a6:8c:08
            ONBOOT = yes
            IPv4_MODE = BACKWARDS_COMPATIBLE
            BOOTPROTO = static
            IPADDR = 10.20.87.154
            """
        )
        nc = self._get_NicConfigurator(config)
        self.assertEqual(
            {
                "NIC1": {
                    "match": {"macaddress": "00:50:56:a6:8c:08"},
                    "wakeonlan": True,
                    "dhcp4": False,
                    "addresses": ["10.20.87.154/32"],
                    "set-name": "NIC1",
                }
            },
            nc.generate(),
        )

    def test_static_nic_without_ipv6_netmask(self):
        """netmask is mandatory for static ipv6 configuration."""
        config = textwrap.dedent(
            """\
            [NETWORK]
            NETWORKING = yes
            BOOTPROTO = dhcp
            HOSTNAME = myhost1
            DOMAINNAME = eng.vmware.com

            [NIC-CONFIG]
            NICS = NIC1

            [NIC1]
            MACADDR = 00:50:56:a6:8c:08
            ONBOOT = yes
            IPv4_MODE = BACKWARDS_COMPATIBLE
            BOOTPROTO = static
            IPADDR = 10.20.87.154
            IPv6ADDR|1 = fc00:10:20:87::154
            """
        )
        nc = self._get_NicConfigurator(config)
        with self.assertRaises(ValueError):
            nc.generate()

    def test_non_primary_nic_with_gateway(self):
        """A non primary nic set can have a gateway."""
        config = textwrap.dedent(
            """\
            [NETWORK]
            NETWORKING = yes
            BOOTPROTO = dhcp
            HOSTNAME = myhost1
            DOMAINNAME = eng.vmware.com

            [NIC-CONFIG]
            NICS = NIC1

            [NIC1]
            MACADDR = 00:50:56:a6:8c:08
            ONBOOT = yes
            IPv4_MODE = BACKWARDS_COMPATIBLE
            BOOTPROTO = static
            IPADDR = 10.20.87.154
            NETMASK = 255.255.252.0
            GATEWAY = 10.20.87.253
            """
        )
        nc = self._get_NicConfigurator(config)
        self.assertEqual(
            {
                "NIC1": {
                    "match": {"macaddress": "00:50:56:a6:8c:08"},
                    "wakeonlan": True,
                    "dhcp4": False,
                    "addresses": ["10.20.87.154/22"],
                    "routes": [{"to": "10.20.84.0/22", "via": "10.20.87.253"}],
                    "set-name": "NIC1",
                }
            },
            nc.generate(),
        )

    def test_cust_non_primary_nic_with_gateway_(self):
        """A customer non primary nic set can have a gateway."""
        config = textwrap.dedent(
            """\
            [NETWORK]
            NETWORKING = yes
            BOOTPROTO = dhcp
            HOSTNAME = static-debug-vm
            DOMAINNAME = cluster.local

            [NIC-CONFIG]
            NICS = NIC1

            [NIC1]
            MACADDR = 00:50:56:ac:d1:8a
            ONBOOT = yes
            IPv4_MODE = BACKWARDS_COMPATIBLE
            BOOTPROTO = static
            IPADDR = 100.115.223.75
            NETMASK = 255.255.255.0
            GATEWAY = 100.115.223.254


            [DNS]
            DNSFROMDHCP=no

            NAMESERVER|1 = 8.8.8.8

            [DATETIME]
            UTC = yes
            """
        )
        nc = self._get_NicConfigurator(config)
        self.assertEqual(
            {
                "NIC1": {
                    "match": {"macaddress": "00:50:56:ac:d1:8a"},
                    "wakeonlan": True,
                    "dhcp4": False,
                    "addresses": ["100.115.223.75/24"],
                    "routes": [
                        {"to": "100.115.223.0/24", "via": "100.115.223.254"}
                    ],
                    "set-name": "NIC1",
                    "nameservers": {"addresses": ["8.8.8.8"]},
                }
            },
            nc.generate(),
        )

    def test_a_primary_nic_with_gateway(self):
        """A primary nic set can have a gateway."""
        config = textwrap.dedent(
            """\
            [NETWORK]
            NETWORKING = yes
            BOOTPROTO = dhcp
            HOSTNAME = myhost1
            DOMAINNAME = eng.vmware.com

            [NIC-CONFIG]
            NICS = NIC1

            [NIC1]
            MACADDR = 00:50:56:a6:8c:08
            ONBOOT = yes
            IPv4_MODE = BACKWARDS_COMPATIBLE
            BOOTPROTO = static
            IPADDR = 10.20.87.154
            NETMASK = 255.255.252.0
            PRIMARY = true
            GATEWAY = 10.20.87.253
            """
        )
        nc = self._get_NicConfigurator(config)
        self.assertEqual(
            {
                "NIC1": {
                    "match": {"macaddress": "00:50:56:a6:8c:08"},
                    "wakeonlan": True,
                    "dhcp4": False,
                    "addresses": ["10.20.87.154/22"],
                    "routes": [{"to": "0.0.0.0/0", "via": "10.20.87.253"}],
                    "set-name": "NIC1",
                }
            },
            nc.generate(),
        )

    def test_meta_data(self):
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")
        conf = Config(cf)
        self.assertIsNone(conf.meta_data_name)
        cf._insertKey("CLOUDINIT|METADATA", "test-metadata")
        conf = Config(cf)
        self.assertEqual("test-metadata", conf.meta_data_name)

    def test_user_data(self):
        cf = ConfigFile("tests/data/vmware/cust-dhcp-2nic.cfg")
        conf = Config(cf)
        self.assertIsNone(conf.user_data_name)
        cf._insertKey("CLOUDINIT|USERDATA", "test-userdata")
        conf = Config(cf)
        self.assertEqual("test-userdata", conf.user_data_name)



"""Integration tests for CLI functionality

These would be for behavior manually invoked by user from the command line
"""

import pytest

from tests.integration_tests.instances import IntegrationInstance
from tests.integration_tests.integration_settings import PLATFORM
from tests.integration_tests.releases import CURRENT_RELEASE

VALID_USER_DATA = """\
#cloud-config
runcmd:
  - echo 'hi' > /var/tmp/test
"""

INVALID_USER_DATA_HEADER = """\
runcmd:
  - echo 'hi' > /var/tmp/test
"""

FAILING_USER_DATA = """\
#cloud-config
bootcmd:
  - exit 1
runcmd:
  - exit 1
"""

# The '-' in 'hashed-password' fails schema validation
INVALID_USER_DATA_SCHEMA = """\
#cloud-config
users:
  - default
  - name: newsuper
    gecos: Big Stuff
    groups: users, admin
    sudo: ALL=(ALL) NOPASSWD:ALL
    hashed-password: asdfasdf
    shell: /bin/bash
    lock_passwd: true
"""


@pytest.mark.user_data(VALID_USER_DATA)
class TestValidUserData:
    def test_schema_status(self, class_client: IntegrationInstance):
        """Test `cloud-init schema` with valid userdata.

        PR #575
        """
        result = class_client.execute("cloud-init schema --system")
        assert result.ok
        assert "Valid schema user-data" in result.stdout.strip()
        result = class_client.execute("cloud-init status --long")
        assert 0 == result.return_code, (
            f"Unexpected exit {result.return_code} from cloud-init status:"
            f" {result}"
        )

    def test_modules_init(self, class_client: IntegrationInstance):
        for mode in ("init", "config", "final"):
            result = class_client.execute(f"cloud-init modules --mode {mode}")
            assert result.ok
            assert f"'modules:{mode}'" in result.stdout.strip()


@pytest.mark.skipif(
    PLATFORM == "qemu", reason="QEMU only supports #cloud-config userdata"
)
@pytest.mark.user_data(INVALID_USER_DATA_HEADER)
def test_invalid_userdata(client: IntegrationInstance):
    """Test `cloud-init schema` with invalid userdata.

    PR #575
    """
    result = client.execute("cloud-init schema --system")
    assert not result.ok
    assert "Cloud config schema errors" in result.stderr
    assert (
        "Expected first line to be one of: #!, ## template: jinja,"
        " #cloud-boothook, #cloud-config" in result.stderr
    )
    result = client.execute("cloud-init status --long")
    if CURRENT_RELEASE.series in ("focal", "jammy", "lunar", "mantic"):
        return_code = 0  # Stable releases don't change exit code behavior
    else:
        return_code = 2  # 23.4 and later will exit 2 on warnings
    assert (
        return_code == result.return_code
    ), f"Unexpected exit code {result.return_code}"


@pytest.mark.user_data(INVALID_USER_DATA_SCHEMA)
def test_invalid_userdata_schema(client: IntegrationInstance):
    """Test invalid schema represented as Warnings, not fatal

    PR #1175
    """
    result = client.execute("cloud-init status --long")
    if CURRENT_RELEASE.series in ("focal", "jammy", "lunar", "mantic"):
        return_code = 0  # Stable releases don't change exit code behavior
    else:
        return_code = 2  # 23.4 and later will exit 2 on warnings
    assert (
        return_code == result.return_code
    ), f"Unexpected exit code {result.return_code}"
    log = client.read_from_file("/var/log/cloud-init.log")
    warning = (
        "[WARNING]: cloud-config failed schema validation! "
        "You may run 'sudo cloud-init schema --system' to check the details."
    )
    assert warning in log
    assert "asdfasdf" not in log


@pytest.mark.user_data(FAILING_USER_DATA)
def test_failing_userdata_modules_exit_codes(client: IntegrationInstance):
    """Test failing in modules representd in exit status.

    To ensure we don't miss any errors or warnings if a service happens
    to be restarted, any further module invocations will exit with error
    on the same boot if a previous invocation exited with error.

    In this test, both bootcmd and runcmd will exit with error the first time.
    The second time, runcmd will run cleanly, but still exit with error.
    Since bootcmd runs in init timeframe, and runcmd runs in final timeframe,
    expect error from those two modes.
    """
    for mode in ("init", "config", "final"):
        result = client.execute(f"cloud-init modules --mode {mode}")
        assert result.ok if mode == "config" else result.failed
        assert f"'modules:{mode}'" in result.stdout.strip()



# This file is part of cloud-init. See LICENSE file for license information.
from logging import DEBUG, WARNING
from pathlib import Path

import pytest

from cloudinit.config.cc_final_message import handle
from tests.unittests.util import get_cloud


class TestHandle:
    # TODO: Expand these tests to cover full functionality; currently they only
    # cover the logic around how the boot-finished file is written (and not its
    # contents).

    @pytest.mark.parametrize(
        "instance_dir_exists,file_is_written,expected_log_substring",
        [
            (True, True, None),
            (False, False, "Failed to write boot finished file "),
        ],
    )
    def test_boot_finished_written(
        self,
        instance_dir_exists,
        file_is_written,
        expected_log_substring,
        caplog,
        paths,
        tmpdir,
    ):
        instance_dir = Path(paths.get_ipath_cur())
        if instance_dir_exists:
            instance_dir.mkdir()
        boot_finished = instance_dir / "boot-finished"

        m_cloud = get_cloud(paths=paths)
        handle(None, {}, m_cloud, [])

        # We should not change the status of the instance directory
        assert instance_dir_exists == instance_dir.exists()
        assert file_is_written == boot_finished.exists()

        if expected_log_substring:
            assert expected_log_substring in caplog.text

    @pytest.mark.parametrize(
        "dsname,datasource_list,expected_log,log_level",
        [
            ("None", ["None"], "Used fallback datasource", DEBUG),
            ("None", ["LXD", "None"], "Used fallback datasource", WARNING),
            ("LXD", ["LXD", "None"], None, DEBUG),
        ],
    )
    def test_only_warn_when_datasourcenone_is_fallback_in_datasource_list(
        self,
        dsname,
        datasource_list,
        expected_log,
        log_level,
        caplog,
        paths,
    ):
        """Only warn when None is a fallback in multi-item datasource_list.

        It is not a warning when datasource_list: [ None ] is configured.
        """
        m_cloud = get_cloud(paths=paths)
        m_cloud.datasource.dsname = dsname
        Path(paths.get_ipath_cur()).mkdir()
        with caplog.at_level(log_level):
            handle(None, {}, m_cloud, [])

        # We should not change the status of the instance directory
        if expected_log:
            assert expected_log in caplog.text
        else:
            assert "Used fallback datasource" not in caplog.text



# Copyright (C) 2012 Canonical Ltd.
# Copyright (C) 2012 Yahoo! Inc.
#
# Author: Scott Moser <scott.moser@canonical.com>
# Author: Joshua Harlow <harlowja@yahoo-inc.com>
#
# This file is part of cloud-init. See LICENSE file for license information.
import contextlib
import inspect
import logging
import signal
import sys
import threading
import types
from io import StringIO
from typing import Callable, Dict, Final, NamedTuple, Union

from cloudinit import version as vr
from cloudinit.log import log_util

LOG = logging.getLogger(__name__)

SIG_MESSAGE: Final = "Cloud-init {} received {}, exiting\n"
BACK_FRAME_TRACE_DEPTH: Final = 3
SIGNALS: Final[Dict[int, str]] = {
    signal.SIGINT: "Cloud-init %(version)s received SIGINT, exiting",
    signal.SIGTERM: "Cloud-init %(version)s received SIGTERM, exiting",
    signal.SIGABRT: "Cloud-init %(version)s received SIGABRT, exiting",
}


class ExitBehavior(NamedTuple):
    exit_code: int
    log_level: int


SIGNAL_EXIT_BEHAVIOR_CRASH: Final = ExitBehavior(1, logging.ERROR)
SIGNAL_EXIT_BEHAVIOR_QUIET: Final = ExitBehavior(0, logging.INFO)
_SIGNAL_EXIT_BEHAVIOR = SIGNAL_EXIT_BEHAVIOR_CRASH
_SUSPEND_WRITE_LOCK = threading.RLock()


def inspect_handler(sig: Union[int, Callable, None]) -> None:
    """inspect_handler() logs signal handler state"""
    if callable(sig):
        # only produce a log when the signal handler isn't in the expected
        # default state
        if not isinstance(sig, types.BuiltinFunctionType):
            LOG.info("Signal state [%s] - previously custom handler.", sig)
    elif sig == signal.SIG_IGN:
        LOG.info("Signal state [SIG_IGN] - previously ignored.")
    elif sig is None:
        LOG.info("Signal state [None] - previously not installed from Python.")
    elif sig == signal.SIG_DFL:
        LOG.info(
            "Signal state [%s] - default way of handling signal was "
            "previously in use.",
            sig,
        )
    else:
        # this should never happen, unless something in Python changes
        # https://docs.python.org/3/library/signal.html#signal.getsignal
        LOG.warning("Signal state [%s(%s)] - unknown", type(sig), sig)


def _pprint_frame(frame, depth, max_depth, contents):
    if depth > max_depth or not frame:
        return
    frame_info = inspect.getframeinfo(frame)
    prefix = " " * (depth * 2)
    contents.write("%sFilename: %s\n" % (prefix, frame_info.filename))
    contents.write("%sFunction: %s\n" % (prefix, frame_info.function))
    contents.write("%sLine number: %s\n" % (prefix, frame_info.lineno))
    _pprint_frame(frame.f_back, depth + 1, max_depth, contents)


def _handle_exit(signum, frame):
    # in practice we always receive a Signals object but int is possible
    name = signum.name if isinstance(signum, signal.Signals) else signum
    contents = StringIO(SIG_MESSAGE.format(vr.version_string(), name))
    _pprint_frame(frame, 1, BACK_FRAME_TRACE_DEPTH, contents)
    log_util.multi_log(
        contents.getvalue(), log=LOG, log_level=_SIGNAL_EXIT_BEHAVIOR.log_level
    )
    sys.exit(_SIGNAL_EXIT_BEHAVIOR.exit_code)


def attach_handlers():
    """attach cloud-init's handlers"""
    sigs_attached = 0
    for signum in SIGNALS.keys():
        inspect_handler(signal.signal(signum, _handle_exit))
    sigs_attached += len(SIGNALS)
    return sigs_attached


@contextlib.contextmanager
def suspend_crash():
    """suspend_crash() allows signals to be received without exiting 1

    This allow signal handling without a crash where it is expected. The
    call stack is still printed if signal is received during this context, but
    the return code is 0 and no traceback is printed.

    Threadsafe.
    """
    global _SIGNAL_EXIT_BEHAVIOR

    # If multiple threads simultaneously were to modify this
    # global state, this function would not behave as expected.
    with _SUSPEND_WRITE_LOCK:
        _SIGNAL_EXIT_BEHAVIOR = SIGNAL_EXIT_BEHAVIOR_QUIET
        yield
        _SIGNAL_EXIT_BEHAVIOR = SIGNAL_EXIT_BEHAVIOR_CRASH



# This file is part of cloud-init. See LICENSE file for license information.

import logging
from pathlib import Path
from unittest import mock

import pytest
from configobj import ConfigObj

from cloudinit import util
from cloudinit.config import cc_set_hostname
from cloudinit.sources.DataSourceNone import DataSourceNone
from tests.unittests.util import get_cloud

LOG = logging.getLogger(__name__)


def fake_subp(*args, **kwargs):
    if args[0][0] in ["hostname", "hostnamectl"]:
        return None, None
    raise RuntimeError(f"Unexpected subp: {args[0]}")


def conf_parser(conf):
    return dict(ConfigObj(conf.splitlines()))


@pytest.mark.usefixtures("fake_filesystem")
class TestHostname:
    @pytest.fixture(autouse=True)
    def common_mocks(self, mocker):
        mocker.patch(
            "cloudinit.distros.debian.subp.subp", side_effect=fake_subp
        )

    @pytest.mark.parametrize(
        "distro_name,cfg,host_path,parser,expected",
        (
            pytest.param(
                "debian",
                {"hostname": "blah", "fqdn": "blah.example.com"},
                "/etc/hostname",
                lambda x: x,
                "blah",
                id="debian",
            ),
            pytest.param(
                "debian",
                {
                    "hostname": "blah",
                    "prefer_fqdn_over_hostname": True,
                    "fqdn": "blah.example.com",
                },
                "/etc/hostname",
                lambda x: x,
                "blah.example.com",
                id="debian_prefer_fqdn",
            ),
            pytest.param(
                "rhel",
                {"hostname": "blah", "fqdn": "blah.example.com"},
                "/etc/sysconfig/network",
                conf_parser,
                {"HOSTNAME": "blah.example.com"},
                id="rhel",
            ),
            pytest.param(
                "rhel",
                {
                    "hostname": "blah",
                    "prefer_fqdn_over_hostname": False,
                    "fqdn": "blah.example.com",
                },
                "/etc/sysconfig/network",
                conf_parser,
                {"HOSTNAME": "blah"},
                id="rhel_prefer_hostname",
            ),
            pytest.param(
                "sles",
                {"hostname": "blah", "fqdn": "blah.example.com"},
                "/etc/HOSTNAME",
                lambda x: x,
                "blah",
                id="sles",
            ),
        ),
    )
    def test_write_hostname(
        self,
        distro_name,
        cfg,
        host_path,
        parser,
        expected,
        paths,
        mocker,
    ):
        mocker.patch(
            "cloudinit.distros.Distro.uses_systemd", return_value=False
        )
        cc = get_cloud(distro=distro_name, paths=paths, sys_cfg=cfg)
        cc_set_hostname.handle("cc_set_hostname", cfg, cc, [])
        contents = util.load_text_file(host_path).strip()
        assert expected == parser(contents)

    @mock.patch("cloudinit.distros.photon.subp.subp")
    def test_photon_hostname(self, m_subp, paths):
        cfg1 = {
            "hostname": "photon",
            "prefer_fqdn_over_hostname": True,
            "fqdn": "test1.vmware.com",
        }
        cfg2 = {
            "hostname": "photon",
            "prefer_fqdn_over_hostname": False,
            "fqdn": "test2.vmware.com",
        }

        m_subp.return_value = (None, None)
        cc = get_cloud(distro="photon", paths=paths, sys_cfg=cfg1)
        for c in [cfg1, cfg2]:
            cc_set_hostname.handle("cc_set_hostname", c, cc, [])
            print("\n", m_subp.call_args_list)
            if c["prefer_fqdn_over_hostname"]:
                assert [
                    mock.call(
                        ["hostnamectl", "set-hostname", c["fqdn"]],
                        capture=True,
                    )
                ] in m_subp.call_args_list
                assert [
                    mock.call(
                        ["hostnamectl", "set-hostname", c["hostname"]],
                        capture=True,
                    )
                ] not in m_subp.call_args_list
            else:
                assert [
                    mock.call(
                        ["hostnamectl", "set-hostname", c["hostname"]],
                        capture=True,
                    )
                ] in m_subp.call_args_list
                assert [
                    mock.call(
                        ["hostnamectl", "set-hostname", c["fqdn"]],
                        capture=True,
                    )
                ] not in m_subp.call_args_list

    @mock.patch("cloudinit.util.get_hostname", return_value="localhost")
    def test_multiple_calls_skips_unchanged_hostname(
        self, get_hostname, paths, caplog
    ):
        """Only new hostname or fqdn values will generate a hostname call."""
        cc = get_cloud(distro="debian", paths=paths)
        cc_set_hostname.handle(
            "cc_set_hostname", {"hostname": "hostname1.me.com"}, cc, []
        )
        contents = util.load_text_file("/etc/hostname")
        assert "hostname1" == contents.strip()
        cc_set_hostname.handle(
            "cc_set_hostname", {"hostname": "hostname1.me.com"}, cc, []
        )
        assert "No hostname changes. Skipping set_hostname\n" in caplog.text
        cc_set_hostname.handle(
            "cc_set_hostname", {"hostname": "hostname2.me.com"}, cc, []
        )
        contents = util.load_text_file("/etc/hostname")
        assert "hostname2" == contents.strip()
        assert (
            "Non-persistently setting the system hostname to hostname2"
            in caplog.text
        )

    @mock.patch("cloudinit.util.get_hostname", return_value="localhost")
    def test_localhost_default_hostname(self, get_hostname, paths):
        """
        No hostname set. Default value returned is localhost,
        but we shouldn't write it in /etc/hostname
        """
        cc = get_cloud(distro="debian", paths=paths, ds=DataSourceNone)

        util.write_file("/etc/hostname", "")
        cc_set_hostname.handle("cc_set_hostname", {}, cc, [])
        contents = util.load_text_file("/etc/hostname")
        assert "" == contents.strip()

    @mock.patch("cloudinit.util.get_hostname", return_value="localhost")
    def test_localhost_user_given_hostname(self, get_hostname, paths):
        """
        User set hostname is localhost. We should write it in /etc/hostname
        """
        cc = get_cloud(distro="debian", paths=paths, ds=DataSourceNone)

        # user-provided localhost should not be ignored
        util.write_file("/etc/hostname", "")
        cc_set_hostname.handle(
            "cc_set_hostname", {"hostname": "localhost"}, cc, []
        )
        contents = util.load_text_file("/etc/hostname")
        assert "localhost" == contents.strip()

    def test_error_on_distro_set_hostname_errors(self, paths):
        """Raise SetHostnameError on exceptions from distro.set_hostname."""

        def set_hostname_error(hostname, fqdn=None) -> None:
            raise RuntimeError(f"OOPS on: {fqdn}")

        cc = get_cloud(distro="debian", paths=paths)
        cc.distro.set_hostname = set_hostname_error
        with pytest.raises(cc_set_hostname.SetHostnameError) as exc_info:
            cc_set_hostname.handle(
                "somename", {"hostname": "hostname1.me.com"}, cc, []
            )
        assert (
            "Failed to set the hostname to hostname1.me.com (hostname1):"
            " OOPS on: hostname1.me.com" == str(exc_info.value)
        )

    def test_ignore_empty_previous_artifact_file(self, paths):
        cfg = {
            "hostname": "blah",
            "fqdn": "blah.blah.blah.yahoo.com",
        }
        cc = get_cloud(distro="debian", paths=paths)
        prev_fn = Path(cc.get_cpath("data")) / "set-hostname"
        prev_fn.parent.mkdir()
        prev_fn.touch()
        cc_set_hostname.handle("cc_set_hostname", cfg, cc, [])
        contents = util.load_text_file("/etc/hostname")
        assert "blah" == contents.strip()

    @pytest.mark.parametrize(
        "distro_name",
        (
            "debian",
            "arch",
            "alpine",
            "gentoo",
            "photon",
            "rhel",
        ),
    )
    def test_create_hostname_file_false(self, distro_name, paths):
        cfg = {
            "hostname": "foo",
            "fqdn": "foo.blah.yahoo.com",
            "create_hostname_file": False,
        }
        cc = get_cloud(distro=distro_name, paths=paths)
        cc_set_hostname.handle("cc_set_hostname", cfg, cc, [])
        with pytest.raises(FileNotFoundError):
            util.load_text_file("/etc/hostname")



# This file is part of cloud-init. See LICENSE file for license information.
"""A module for common socket helpers."""
import logging
import os
import socket
import sys
from contextlib import suppress

from cloudinit import performance
from cloudinit.settings import DEFAULT_RUN_DIR

LOG = logging.getLogger(__name__)


def sd_notify(message: str):
    """Send a sd_notify message.

    :param message: sd-notify message (must be valid ascii)
    """
    socket_path = os.environ.get("NOTIFY_SOCKET", "")

    if not socket_path:
        # not running under systemd, no-op
        return

    elif socket_path[0] == "@":
        # abstract
        socket_path.replace("@", "\0", 1)

    # unix domain
    elif socket_path[0] != "/":
        raise OSError("Unsupported socket type")

    with socket.socket(
        socket.AF_UNIX, socket.SOCK_DGRAM | socket.SOCK_CLOEXEC
    ) as sock:
        LOG.info("Sending sd_notify(%s)", str(message))
        sock.connect(socket_path)
        sock.sendall(message.encode("ascii"))


class SocketSync:
    """A two way synchronization protocol over Unix domain sockets."""

    def __init__(self, *names: str):
        """Initialize a synchronization context.

        1) Ensure that the socket directory exists.
        2) Bind a socket for each stage.

        Binding the sockets on initialization allows receipt of stage
        "start" notifications prior to the cloud-init stage being ready to
        start.

        :param names: stage names, used as a unique identifiers
        """
        self.stage = ""
        self.remote = ""
        self.first_exception = ""
        self.systemd_exit_code = 0
        self.experienced_any_error = False
        self.sockets = {
            name: socket.socket(
                socket.AF_UNIX, socket.SOCK_DGRAM | socket.SOCK_CLOEXEC
            )
            for name in names
        }
        # ensure the directory exists
        os.makedirs(f"{DEFAULT_RUN_DIR}/share", mode=0o700, exist_ok=True)
        # removing stale sockets and bind
        for name, sock in self.sockets.items():
            socket_path = f"{DEFAULT_RUN_DIR}/share/{name}.sock"
            with suppress(FileNotFoundError):
                os.remove(socket_path)
            sock.bind(socket_path)

    def __call__(self, stage: str):
        """Set the stage before entering context.

        This enables the context manager to be initialized separately from
        each stage synchronization.

        :param stage: the name of a stage to synchronize

        Example:
            sync = SocketSync("stage 1", "stage 2"):
            with sync("stage 1"):
                pass
            with sync("stage 2"):
                pass
        """
        if stage not in self.sockets:
            raise ValueError(f"Invalid stage name: {stage}")
        self.stage = stage
        return self

    def __enter__(self):
        """Wait until a message has been received on this stage's socket.

        Once the message has been received, enter the context.
        """
        if os.isatty(sys.stdin.fileno()):
            LOG.info(
                "Stdin is a tty, so skipping stage synchronization protocol"
            )
            return
        self.systemd_exit_code = 0
        sd_notify(
            "STATUS=Waiting on external services to "
            f"complete before starting the {self.stage} stage."
        )
        # block until init system sends us data
        # the first value returned contains a message from the init system
        #     (should be "start")
        # the second value contains the path to a unix socket on which to
        #     reply, which is expected to be /path/to/{self.stage}-return.sock
        sock = self.sockets[self.stage]
        with performance.Timed(f"Waiting to start stage {self.stage}"):
            chunk, self.remote = sock.recvfrom(5)

        if b"start" != chunk:
            # The protocol expects to receive a command "start"
            self.__exit__(None, None, None)
            raise ValueError(f"Received invalid message: [{str(chunk)}]")
        elif f"{DEFAULT_RUN_DIR}/share/{self.stage}-return.sock" != str(
            self.remote
        ):
            # assert that the return path is in a directory with appropriate
            # permissions
            self.__exit__(None, None, None)
            raise ValueError(f"Unexpected path to unix socket: {self.remote}")

        sd_notify(f"STATUS=Running ({self.stage} stage)")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Notify the socket that this stage is complete."""
        message = f"Completed socket interaction for boot stage {self.stage}"
        if exc_type:
            # handle exception thrown in context
            self.systemd_exit_code = 1
            self.experienced_any_error = True
            status = f"{repr(exc_val)} in {exc_tb.tb_frame}"
            message = (
                'fatal error, run "systemctl status cloud-init-main.service" '
                'and "cloud-init status --long" for more details'
            )
            if not self.first_exception:
                self.first_exception = status
            LOG.fatal(status)
            sd_notify(f"STATUS={status}")

        self.experienced_any_error = self.experienced_any_error or bool(
            self.systemd_exit_code
        )
        sock = self.sockets[self.stage]
        sock.connect(self.remote)

        # the returned message will be executed in a subshell
        # hardcode this message rather than sending a more informative message
        # to avoid having to sanitize inputs (to prevent escaping the shell)
        sock.sendall(
            f"echo '{message}'; exit {self.systemd_exit_code};".encode()
        )
        sock.close()

        # suppress exception - the exception was logged and the init system
        # notified of stage completion (and the exception received as a status
        # message). Raising an exception would block the rest of boot, so carry
        # on in a degraded state.
        return True



import logging
import re

import pytest

from tests.integration_tests.clouds import (
    IntegrationCloud,
    IntegrationInstance,
)
from tests.integration_tests.integration_settings import PLATFORM
from tests.integration_tests.util import verify_clean_boot, verify_clean_log

logger = logging.getLogger(__name__)

USER_DATA = """\
#cloud-config
drivers:
    nvidia:
        license-accepted: true
"""


@pytest.mark.skipif(PLATFORM != "oci", reason="Test is OCI specific")
def test_ubuntu_drivers_installed(session_cloud: IntegrationCloud):
    """
    Test the installation of NVIDIA drivers on an OCI instance.

    This test checks that the ubuntu-drivers module installs NVIDIA drivers
    as expected on an OCI instance.

    This test launches its own instance so that it can ensure that a GPU
    instance type is used. The "VM.GPU.A10.1" instance type is used because
    it is the most widely available GPU instance type on OCI.

    Additionally, in case there is limited availability of GPU instances, this
    test launches an instance outside the normal context manager used in other
    tests. This is so that if the instance fails to launch, the test can be
    marked as xfail rather than just failing.

    Test Steps:
    1. Launch a GPU instance with the user data that installs NVIDIA drivers.
    2. Verify that the cloud-init log is clean.
    3. Verify that the instance boots cleanly.
    4. Verify that the NVIDIA drivers are installed as expected.
    """
    try:
        client = session_cloud.launch(
            launch_kwargs={"instance_type": "VM.GPU.A10.1"},
            user_data=USER_DATA,
        )
    except Exception as e:
        pytest.xfail(f"Instance launch failed: {e}")

    try:
        _do_test(client)
    finally:
        client.destroy()


def _do_test(client: IntegrationInstance):
    """Performs above test steps on the given client."""
    log = client.read_from_file("/var/log/cloud-init.log")
    verify_clean_log(log)
    verify_clean_boot(client)
    assert 1 == log.count(
        "Installing and activating NVIDIA drivers "
        "(nvidia/license-accepted=True, version=latest)"
    )
    result = client.execute("dpkg -l | grep nvidia")
    assert result.ok, "No nvidia packages found"
    assert re.search(
        r"ii\s+nvidia.*-\d+-server", result.stdout
    ), f"Did not find specific nvidia driver packages in: {result.stdout}"



# This file is part of cloud-init. See LICENSE file for license information.
import logging

import pytest

from cloudinit.distros import fetch
from cloudinit.subp import SubpResult


class TestPackageCommand:
    @pytest.mark.parametrize("snap_available", (True, False))
    def test_package_command_only_refresh_snap_when_available(
        self, snap_available, mocker
    ):
        """Avoid calls to snap refresh when snap command not available."""
        m_snap_available = mocker.patch(
            "cloudinit.distros.ubuntu.Snap.available",
            return_value=snap_available,
        )
        m_snap_upgrade_packages = mocker.patch(
            "cloudinit.distros.ubuntu.Snap.upgrade_packages",
            return_value=snap_available,
        )
        m_apt_run_package_command = mocker.patch(
            "cloudinit.distros.package_management.apt.Apt.run_package_command",
        )
        cls = fetch("ubuntu")
        distro = cls("ubuntu", {}, None)
        distro.package_command("upgrade")
        m_apt_run_package_command.assert_called_once_with("upgrade")
        m_snap_available.assert_called_once()
        if snap_available:
            m_snap_upgrade_packages.assert_called_once()
        else:
            m_snap_upgrade_packages.assert_not_called()

    @pytest.mark.parametrize(
        "subp_side_effect,expected_log",
        (
            pytest.param(
                [
                    SubpResult(
                        stdout='{"refresh": {"hold": "forever"}}', stderr=None
                    )
                ],
                "Skipping snap refresh because refresh.hold is set to"
                " 'forever'",
                id="skip_snap_refresh_due_to_global_hold_forever",
            ),
            pytest.param(
                [
                    SubpResult(
                        stdout=(
                            '{"refresh": {"hold":'
                            ' "2024-07-08T15:38:20-06:00"}}'
                        ),
                        stderr=None,
                    ),
                    SubpResult(stdout="All snaps up to date.", stderr=""),
                ],
                "",
                id="perform_snap_refresh_due_to_temporary_global_hold",
            ),
            pytest.param(
                [
                    SubpResult(
                        stdout="{}",
                        stderr=(
                            'error: snap "core" has no "refresh.hold" '
                            "configuration option"
                        ),
                    ),
                    SubpResult(stdout="All snaps up to date.", stderr=""),
                ],
                "",
                id="snap_refresh_performed_when_no_global_hold_is_set",
            ),
        ),
    )
    def test_package_command_avoids_snap_refresh_when_refresh_hold_is_forever(
        self, subp_side_effect, expected_log, caplog, mocker
    ):
        """Do not call snap refresh when snap refresh.hold is forever.

        This indicates an environment where snaps refreshes are not preferred
        for whatever reason.
        """
        m_snap_available = mocker.patch(
            "cloudinit.distros.ubuntu.Snap.available",
            return_value=True,
        )
        m_subp = mocker.patch(
            "cloudinit.subp.subp",
            side_effect=subp_side_effect,
        )
        m_apt_run_package_command = mocker.patch(
            "cloudinit.distros.package_management.apt.Apt.run_package_command",
        )
        cls = fetch("ubuntu")
        distro = cls("ubuntu", {}, None)
        with caplog.at_level(logging.INFO):
            distro.package_command("upgrade")
        m_apt_run_package_command.assert_called_once_with("upgrade")
        m_snap_available.assert_called_once()
        expected_calls = [mocker.call(["snap", "get", "system", "-d"])]
        if expected_log:
            assert expected_log in caplog.text
        else:
            expected_calls.append(mocker.call(["snap", "refresh"]))
        assert m_subp.call_args_list == expected_calls



# This file is part of cloud-init. See LICENSE file for license information.

import json
import os
import stat

from cloudinit import atomic_helper


class TestAtomicHelper:
    def test_basic_usage(self, tmp_path):
        """write_file takes bytes if no omode."""
        path = tmp_path / "test_basic_usage"
        contents = b"Hey there\n"
        atomic_helper.write_file(path, contents)
        self.check_file(path, contents)

    def test_string(self, tmp_path):
        """write_file can take a string with mode w."""
        path = tmp_path / "test_string"
        contents = "Hey there\n"
        atomic_helper.write_file(path, contents, omode="w")
        self.check_file(path, contents, omode="r")

    def test_file_permissions(self, tmp_path):
        """write_file with mode 400 works correctly."""
        path = tmp_path / "test_file_permissions"
        contents = b"test_file_perms"
        atomic_helper.write_file(path, contents, mode=0o400)
        self.check_file(path, contents, perms=0o400)

    def test_file_preserve_permissions(self, tmp_path):
        """create a file with mode 700, then write_file with mode 644."""
        path = tmp_path / "test_file_preserve_permissions"
        contents = b"test_file_perms"
        with open(path, mode="wb") as f:
            f.write(b"test file preserve permissions")
            os.chmod(f.name, 0o700)
            atomic_helper.write_file(path, contents, preserve_mode=True)
            self.check_file(path, contents, perms=0o700)

    def test_write_json(self, tmp_path):
        """write_json output is readable json."""
        path = tmp_path / "test_write_json"
        data = {"key1": "value1", "key2": ["i1", "i2"]}
        atomic_helper.write_json(path, data)
        with open(path, "r") as fp:
            found = json.load(fp)
        assert data == found
        self.check_perms(path, 0o644)

    def check_file(self, path, content, omode=None, perms=0o644):
        if omode is None:
            omode = "rb"
        assert os.path.exists(path)
        assert os.path.isfile(path)
        with open(path, omode) as fp:
            found = fp.read()
            assert content == found
        self.check_perms(path, perms)

    def check_perms(self, path, perms):
        file_stat = os.stat(path)
        assert perms == stat.S_IMODE(file_stat.st_mode)

    def test_write_file_ensure_dirs(self, tmp_path):
        path = tmp_path / "ensure_dirs" / "ensure/dir"
        contents = b"Hey there\n"
        atomic_helper.write_file(path, contents)
        self.check_file(path, contents)



import json
import logging
import multiprocessing
import os
import re
import time
from collections import namedtuple
from contextlib import contextmanager
from functools import lru_cache
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Set, Union

import pytest

from cloudinit.subp import subp
from tests.integration_tests.decorators import retry
from tests.integration_tests.integration_settings import (
    OS_IMAGE_TYPE,
    PLATFORM,
)
from tests.integration_tests.releases import CURRENT_RELEASE, NOBLE

LOG = logging.getLogger("integration_testing.util")

if TYPE_CHECKING:
    # instances.py has imports util.py, so avoid circular import
    from tests.integration_tests.instances import IntegrationInstance

log = logging.getLogger("integration_testing")
key_pair = namedtuple("key_pair", "public_key private_key")

ASSETS_DIR = Path("tests/integration_tests/assets")
KEY_PATH = ASSETS_DIR / "keys"

HAS_CONSOLE_LOG = PLATFORM in (
    "ec2",
    "lxd_container",
    "oci",
    "openstack",
    "qemu",
)


def verify_ordered_items_in_text(to_verify: list, text: str):
    """Assert all items in list appear in order in text.

    Examples:
      verify_ordered_items_in_text(['a', '1'], 'ab1')  # passes
      verify_ordered_items_in_text(['1', 'a'], 'ab1')  # raises AssertionError
    """
    index = 0
    for item in to_verify:
        try:
            matched = re.search(item, text[index:])
        except re.error:
            matched = re.search(re.escape(item), text[index:])
        assert matched, "Expected item not found: '{}'".format(item)
        index = matched.start()


def _format_found(header: str, items: list) -> str:
    """Helper function to format assertion message"""

    # do nothing, allows this formatter to be "stackable"
    if not items:
        return ""

    # if only one error put the header and the error message on a single line
    if 1 == len(items):
        return f"\n{header}: {items.pop(0)}"

    # otherwise make a list after header
    else:
        return f"\n{header}:\n\t- " + "\n\t- ".join(items)


def verify_clean_boot(
    instance: "IntegrationInstance",
    *,
    ignore_deprecations: Optional[Union[List[str], bool]] = None,
    ignore_warnings: Optional[Union[List[str], bool]] = None,
    ignore_errors: Optional[Union[List[str], bool]] = None,
    ignore_tracebacks: Optional[Union[List[str], bool]] = None,
    require_deprecations: Optional[list] = None,
    require_warnings: Optional[list] = None,
    require_errors: Optional[list] = None,
):
    """Raise exception if the client experienced unexpected conditions.

    Fail when a required error, warning, or deprecation isn't found.

    Platform-specific conditions are defined in this function.

    This function is similar to verify_clean_log, hence the similar name.

    differences from verify_clean_log:

    - more expressive syntax
    - extensible (can be easily extended for other log levels)
    - less resource intensive (no log copying required)
    - nice error formatting

    :param instance: test instance
    :param ignored_deprecations: list of deprecation messages to ignore, or
        true to ignore all
    :param ignored_warnings: list of expected warnings to ignore, or true to
        ignore all
    :param ignored_errors: list of expected errors to ignore, or true to ignore
        all
    :param require_deprecations: list of expected deprecation messages to
        require
    :param require_warnings: list of expected warning messages to require
    :param require_errors: list of expected error messages to require
    """

    def append_or_create_list(
        maybe_list: Optional[Union[List[str], bool]], value: str
    ) -> Optional[Union[List[str], bool]]:
        """handle multiple types"""
        if isinstance(maybe_list, list):
            maybe_list.append(value)
        elif maybe_list is True:
            return True  # Ignoring all texts, so no need to append.
        elif maybe_list in (None, False):
            maybe_list = [value]
        return maybe_list

    if ignore_tracebacks is None:
        ignore_tracebacks = []
    # Define exceptions by matrix of platform and Ubuntu release
    if "azure" == PLATFORM:
        # Consistently on all Azure launches:
        ignore_warnings = append_or_create_list(
            ignore_warnings, "No lease found; using default endpoint"
        )
    elif "lxd_vm" == PLATFORM:
        # Ubuntu lxd storage
        ignore_warnings = append_or_create_list(
            ignore_warnings, "thinpool by default on Ubuntu due to LP #1982780"
        )
        ignore_warnings = append_or_create_list(
            ignore_warnings,
            "Could not match supplied host pattern, ignoring:",
        )
    elif "oci" == PLATFORM:
        # LP: #1842752
        ignore_errors = append_or_create_list(
            ignore_warnings, "Stderr: RTNETLINK answers: File exists"
        )
        if isinstance(ignore_tracebacks, list):
            ignore_tracebacks.append("Stderr: RTNETLINK answers: File exists")
        # Ubuntu lxd storage
        ignore_warnings = append_or_create_list(
            ignore_warnings, "thinpool by default on Ubuntu due to LP #1982780"
        )
        # LP: #1833446
        ignore_warnings = append_or_create_list(
            ignore_warnings,
            "UrlError: 404 Client Error: Not Found for url: "
            "http://169.254.169.254/latest/meta-data/",
        )
        if isinstance(ignore_tracebacks, list):
            ignore_tracebacks.append(
                "UrlError: 404 Client Error: Not Found for url: "
                "http://169.254.169.254/latest/meta-data/"
            )
        # Oracle has a file in /etc/cloud/cloud.cfg.d that contains
        # users:
        # - default
        # - name: opc
        #   ssh_redirect_user: true
        # This can trigger a warning about opc having no public key
        ignore_warnings = append_or_create_list(
            ignore_warnings,
            "Unable to disable SSH logins for opc given ssh_redirect_user",
        )
    # Preserve platform-specific tracebacks expected
    _verify_clean_boot(
        instance,
        ignore_deprecations=ignore_deprecations,
        ignore_warnings=ignore_warnings,
        ignore_errors=ignore_errors,
        ignore_tracebacks=ignore_tracebacks,
        require_deprecations=require_deprecations,
        require_warnings=require_warnings,
        require_errors=require_errors,
    )


def _verify_clean_boot(
    instance: "IntegrationInstance",
    ignore_deprecations: Optional[Union[List[str], bool]] = None,
    ignore_warnings: Optional[Union[List[str], bool]] = None,
    ignore_errors: Optional[Union[List[str], bool]] = None,
    ignore_tracebacks: Optional[Union[List[str], bool]] = None,
    require_deprecations: Optional[list] = None,
    require_warnings: Optional[list] = None,
    require_errors: Optional[list] = None,
):
    ignore_deprecations = ignore_deprecations or []
    ignore_errors = ignore_errors or []
    ignore_warnings = ignore_warnings or []
    require_deprecations = require_deprecations or []
    require_errors = require_errors or []
    require_warnings = require_warnings or []
    status = json.loads(instance.execute("cloud-init status --format=json"))

    unexpected_deprecations = set()
    unexpected_errors = set()
    unexpected_warnings = set()

    required_deprecations_found = set()
    required_warnings_found = set()
    required_errors_found = set()

    for current_error in status["errors"]:

        # check for required errors
        for expected in require_errors:
            if expected in current_error:
                required_errors_found.add(expected)

        if ignore_errors is True:
            continue
        # check for unexpected errors
        for expected in [*ignore_errors, *require_errors]:
            if expected in current_error:
                break
        else:
            unexpected_errors.add(current_error)

    # check for unexpected warnings
    for current_warning in status["recoverable_errors"].get("WARNING", []):

        # check for required warnings
        for expected in require_warnings:
            if expected in current_warning:
                required_warnings_found.add(expected)

        if ignore_warnings is True:
            continue
        # check for unexpected warnings
        for expected in [*ignore_warnings, *require_warnings]:
            if expected in current_warning:
                break
        else:
            unexpected_warnings.add(current_warning)

    # check for unexpected deprecations
    for current_deprecation in status["recoverable_errors"].get(
        "DEPRECATED", []
    ):

        # check for required deprecations
        for expected in require_deprecations:
            if expected in current_deprecation:
                required_deprecations_found.add(expected)

        if ignore_deprecations is True:
            continue
        # check for unexpected deprecations
        for expected in [*ignore_deprecations, *require_deprecations]:
            if expected in current_deprecation:
                break
        else:
            unexpected_deprecations.add(current_deprecation)

    required_errors_not_found = set(require_errors) - required_errors_found
    required_warnings_not_found = (
        set(require_warnings) - required_warnings_found
    )
    required_deprecations_not_found = (
        set(require_deprecations) - required_deprecations_found
    )

    # ordered from most severe to least severe
    # this allows the first message printed to be the most significant
    errors = [
        *unexpected_errors,
        *required_errors_not_found,
        *unexpected_warnings,
        *required_warnings_not_found,
        *unexpected_deprecations,
        *required_deprecations_not_found,
    ]
    if errors:
        message = ""
        # if there is only one message, don't include the generic header
        # so that the user can read the exact message in the pytest summary
        if len(errors) > 1:
            # more than one error, so include a generic message
            message += "Unexpected warnings or errors found"

        # errors are probably more important, order them first
        message += _format_found(
            "Found unexpected errors", list(unexpected_errors)
        )
        message += _format_found(
            "Required errors not found", list(required_errors_not_found)
        )
        message += _format_found(
            "Found unexpected warnings", list(unexpected_warnings)
        )
        message += _format_found(
            "Required warnings not found", list(required_warnings_not_found)
        )
        message += _format_found(
            "Found unexpected deprecations", list(unexpected_deprecations)
        )
        message += _format_found(
            "Required deprecations not found",
            list(required_deprecations_not_found),
        )
        assert not errors, message

    # assert no unexpected Tracebacks
    if ignore_tracebacks is not True:
        expected_traceback_count = 0
        traceback_count = int(
            instance.execute(
                "grep --count Traceback /var/log/cloud-init.log"
            ).stdout.strip()
        )
        if ignore_tracebacks:
            for expected_traceback in ignore_tracebacks:
                expected_traceback_count += int(
                    instance.execute(
                        f"grep --count '{expected_traceback}'"
                        " /var/log/cloud-init.log"
                    ).stdout.strip()
                )
        assert expected_traceback_count == traceback_count, (
            f"{traceback_count - expected_traceback_count} unexpected "
            "traceback(s) found in /var/log/cloud-init.log"
        )

    # check that the return code of cloud-init status is expected
    if not any(
        [
            ignore_deprecations,
            ignore_warnings,
            ignore_errors,
            ignore_tracebacks,
            require_deprecations,
            require_warnings,
            require_errors,
        ]
    ):
        # make sure that return code is 0 when all was good
        out = instance.execute("cloud-init status --long")
        assert out.ok, (
            "Unexpected non-zero return code from "
            f"`cloud-init status`: {out.return_code}\nstdout: "
            f"{out.stdout}\nstderr: {out.stderr}"
        )
    elif any(
        [
            ignore_deprecations,
            ignore_warnings,
            ignore_errors,
            ignore_tracebacks,
        ]
    ):
        # Ignore is optional, so we can't be sure whether we will got a return
        # code of 0 or 2. This makes this function more useful when writing
        # tests that run on series with different behavior. In this case we
        # cannot be sure whether it will return 0 or 2, but we never want to
        # see a return code of 1, so assert that it isn't 1.
        out = instance.execute("cloud-init status --long")
        assert 1 != out.return_code, (
            f"Unexpected return code from `cloud-init status`. Expected rc=0 "
            f"or rc=2, received rc={out.return_code}\nstdout:\n"
            f"{out.stdout}\nstderr:\n{out.stderr}"
        )
    else:
        # Look at the commit history before making changes here.
        # Behavior on Jenkins, GH CI, and local runs must be considered
        # for all series.
        out = instance.execute("cloud-init status --long")
        rc = 2
        if CURRENT_RELEASE < NOBLE:
            rc = 0
        assert rc == out.return_code, (
            f"Unexpected return code from `cloud-init status`. "
            f"Expected rc={rc}, received rc={out.return_code}\nstdout: "
            f"{out.stdout}\nstderr: {out.stderr}"
        )
    schema = instance.execute("cloud-init schema --system --annotate")
    assert schema.ok, (
        f"Schema validation failed\nstdout:{schema.stdout}"
        f"\nstderr:\n{schema.stderr}"
    )


def verify_clean_log(log: str, ignore_deprecations: bool = True):
    """Assert no unexpected tracebacks or warnings in logs"""
    if ignore_deprecations:
        is_deprecated = re.compile("deprecat", flags=re.IGNORECASE)
        log_lines = log.split("\n")
        log_lines = list(
            filter(lambda line: not is_deprecated.search(line), log_lines)
        )
        log = "\n".join(log_lines)

    error_logs = re.findall("CRITICAL.*", log) + re.findall("ERROR.*", log)
    if error_logs:
        raise AssertionError(
            "Found unexpected errors: %s" % "\n".join(error_logs)
        )
    if re.findall("Cloud-init.*received SIG", log):
        raise AssertionError(
            "Found unexpected signal termination: %s" % "\n".join(error_logs)
        )

    warning_count = log.count("[WARNING]")
    expected_warnings = 0
    traceback_count = log.count("Traceback")
    expected_tracebacks = 0

    warning_texts = [
        # Consistently on all Azure launches:
        # azure.py[WARNING]: No lease found; using default endpoint
        "No lease found; using default endpoint",
        # Ubuntu lxd storage
        "thinpool by default on Ubuntu due to LP #1982780",
        "WARNING]: Could not match supplied host pattern, ignoring:",
    ]
    traceback_texts = []
    if "install canonical-livepatch" in log:
        # Ubuntu Pro Client emits a warning in between installing livepatch
        # and enabling it
        warning_texts.append(
            "canonical-livepatch returned error when checking status"
        )
    if "found network data from DataSourceNone" in log:
        warning_texts.append("Used fallback datasource")
    if "oracle" in log:
        # LP: #1842752
        lease_exists_text = "Stderr: RTNETLINK answers: File exists"
        warning_texts.append(lease_exists_text)
        traceback_texts.append(lease_exists_text)
        # LP: #1833446
        fetch_error_text = (
            "UrlError: 404 Client Error: Not Found for url: "
            "http://169.254.169.254/latest/meta-data/"
        )
        warning_texts.append(fetch_error_text)
        traceback_texts.append(fetch_error_text)
        # Oracle has a file in /etc/cloud/cloud.cfg.d that contains
        # users:
        # - default
        # - name: opc
        #   ssh_redirect_user: true
        # This can trigger a warning about opc having no public key
        warning_texts.append(
            "Unable to disable SSH logins for opc given ssh_redirect_user"
        )

    for warning_text in warning_texts:
        expected_warnings += log.count(warning_text)
    for traceback_text in traceback_texts:
        expected_tracebacks += log.count(traceback_text)

    assert warning_count <= expected_warnings, (
        f"Unexpected warning count != {expected_warnings}. Found: "
        f"{re.findall('WARNING.*', log)}"
    )
    assert traceback_count == expected_tracebacks


def get_inactive_modules(log: str) -> Set[str]:
    matches = re.findall(
        r"Skipping modules '(.*)' because no applicable config is provided.",
        log,
    )
    return set(
        map(
            lambda module: module.strip(),
            chain(*map(lambda match: match.split(","), matches)),
        )
    )


@contextmanager
def emit_dots_on_travis():
    """emit a dot every 60 seconds if running on Travis.

    Travis will kill jobs that don't emit output for a certain amount of time.
    This context manager spins up a background process which will emit a dot to
    stdout every 60 seconds to avoid being killed.

    It should be wrapped selectively around operations that are known to take a
    long time.
    """
    if os.environ.get("TRAVIS") != "true":
        # If we aren't on Travis, don't do anything.
        yield
        return

    def emit_dots():
        while True:
            log.info(".")
            time.sleep(60)

    dot_process = multiprocessing.Process(target=emit_dots)
    dot_process.start()
    try:
        yield
    finally:
        dot_process.terminate()


def get_test_rsa_keypair(key_name: str = "test1") -> key_pair:
    private_key_path = KEY_PATH / "id_rsa.{}".format(key_name)
    public_key_path = KEY_PATH / "id_rsa.{}.pub".format(key_name)
    with public_key_path.open() as public_file:
        public_key = public_file.read()
    with private_key_path.open() as private_file:
        private_key = private_file.read()
    return key_pair(public_key, private_key)


# We're implementing our own here in case cloud-init status --wait
# isn't working correctly (LP: #1966085)
def wait_for_cloud_init(client: "IntegrationInstance", num_retries: int = 30):
    last_exception = None
    for _ in range(num_retries):
        try:
            result = client.execute("cloud-init status")
            if result.return_code in (0, 2) and (
                "running" not in result or "not started" not in result
            ):
                return result
        except Exception as e:
            last_exception = e
        time.sleep(1)
    raise Exception(  # pylint: disable=W0719
        "cloud-init status did not return successfully."
    ) from last_exception


def get_console_log(client: "IntegrationInstance"):
    try:
        console_log = client.instance.console_log()
    except NotImplementedError:
        pytest.skip("NotImplementedError when requesting console log")
    except RuntimeError as e:
        if "open : no such file or directory" in str(e):
            if hasattr(client, "lxc_log"):
                return client.lxc_log
        raise e
    if console_log is None:
        pytest.skip("Console log has not been setup")
    if console_log.lower().startswith("no console output"):
        pytest.fail("no console output")
    if PLATFORM in ("lxd_vm", "lxd_container"):
        # Preserve non empty console log on lxc platforms because
        # lxc console --show-log can be called once and the log is flushed.
        # Multiple calls to --show-log error on "no such file or directory".
        client.lxc_log = console_log  # type: ignore[attr-defined]
    return console_log


@retry(tries=5, delay=1)  # Retry on get_console_log failures
def get_syslog_or_console(client: "IntegrationInstance") -> str:
    """minimal OS_IMAGE_TYPE does not contain rsyslog"""
    if OS_IMAGE_TYPE == "minimal" and HAS_CONSOLE_LOG:
        return get_console_log(client)
    else:
        return client.read_from_file("/var/log/syslog")


@lru_cache()
def lxd_has_nocloud(client: "IntegrationInstance") -> bool:
    # Bionic or Focal may be detected as NoCloud rather than LXD
    lxd_image_metadata = subp(
        ["lxc", "config", "metadata", "show", client.instance.name]
    )
    return "/var/lib/cloud/seed/nocloud" in lxd_image_metadata.stdout


def get_feature_flag_value(client: "IntegrationInstance", key):
    value = client.execute(
        'python3 -c "from cloudinit import features; '
        f'print(features.{key})"'
    ).strip()
    if "NameError" in value:
        raise NameError(f"name '{key}' is not defined")
    return value


def override_kernel_command_line(ds_str: str, instance: "IntegrationInstance"):
    """set the kernel command line and reboot, return after boot done

    This will not work with containers. This is only tested with lxd vms
    but in theory should work on any virtual machine using grub.

    ds_str: the string that will be inserted into /proc/cmdline
    instance: instance to set kernel command line for
    """

    # The final output in /etc/default/grub should be:
    #
    # GRUB_CMDLINE_LINUX="'ds=nocloud;s=http://my-url/'"
    #
    # That ensures that the kernel command line passed into
    # /boot/efi/EFI/ubuntu/grub.cfg will be properly single-quoted
    #
    # Example:
    #
    # linux /boot/vmlinuz-5.15.0-1030-kvm ro 'ds=nocloud;s=http://my-url/'
    #
    # Not doing this will result in a semicolon-delimited ds argument
    # terminating the kernel arguments prematurely.
    assert instance.execute(
        'printf "GRUB_CMDLINE_LINUX=\\"" >> /etc/default/grub'
    ).ok
    assert instance.execute('printf "\'" >> /etc/default/grub').ok
    assert instance.execute(f"printf '{ds_str}' >> /etc/default/grub").ok
    assert instance.execute('printf "\'\\"" >> /etc/default/grub').ok

    # We should probably include non-systemd distros at some point. This should
    # most likely be as simple as updating the output path for grub-mkconfig
    assert instance.execute(
        "grub-mkconfig -o /boot/efi/EFI/ubuntu/grub.cfg"
    ).ok
    assert instance.execute("cloud-init clean --logs").ok
    instance.restart()


def push_and_enable_systemd_unit(
    client: "IntegrationInstance", unit_name: str, content: str
) -> None:
    service_filename = f"/etc/systemd/system/{unit_name}"
    client.write_to_file(service_filename, content)
    client.execute(f"chmod 0644 {service_filename}", use_sudo=True)
    client.execute(f"systemctl enable {unit_name}", use_sudo=True)


def network_wait_logged(log: str) -> bool:
    return (
        "Running command "
        "['systemctl', 'start', 'systemd-networkd-wait-online.service']"
    ) in log



# This file is part of cloud-init. See LICENSE file for license information.
import logging
from abc import ABC, abstractmethod
from functools import partial
from typing import Callable, Dict, Iterable, List, Optional, Type, Union

from cloudinit import subp, util
from cloudinit.net import eni, netplan, network_manager, networkd
from cloudinit.net.netops.iproute2 import Iproute2
from cloudinit.net.network_state import NetworkState

LOG = logging.getLogger(__name__)


class NoActivatorException(Exception):
    pass


def _alter_interface(
    cmd: list, device_name: str, warn_on_stderr: bool = True
) -> bool:
    """Attempt to alter an interface using a command list"""
    return _alter_interface_callable(partial(subp.subp, cmd), warn_on_stderr)


def _alter_interface_callable(
    callable: Callable, warn_on_stderr: bool = True
) -> bool:
    """Attempt to alter an interface using a callable

    this function standardizes logging and response to failure for
    various activators
    """
    try:
        _out, err = callable()
        if len(err):
            log_stderr = LOG.warning if warn_on_stderr else LOG.debug
            log_stderr("Received stderr output: %s", err)
        return True
    except subp.ProcessExecutionError as e:
        util.logexc(LOG, "Running interface command %s failed", e.cmd)
        return False


class NetworkActivator(ABC):
    @staticmethod
    @abstractmethod
    def available(target: Optional[str] = None) -> bool:
        """Return True if activator is available, otherwise return False."""
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def bring_up_interface(device_name: str) -> bool:
        """Bring up interface.

        Return True is successful, otherwise return False
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def bring_down_interface(device_name: str) -> bool:
        """Bring down interface.

        Return True is successful, otherwise return False
        """
        raise NotImplementedError()

    @classmethod
    def bring_up_interfaces(cls, device_names: Iterable[str]) -> bool:
        """Bring up specified list of interfaces.

        Return True is successful, otherwise return False
        """
        return all(cls.bring_up_interface(device) for device in device_names)

    @classmethod
    def bring_up_all_interfaces(cls, network_state: NetworkState) -> bool:
        """Bring up all interfaces.

        Return True is successful, otherwise return False
        """
        return cls.bring_up_interfaces(
            [i["name"] for i in network_state.iter_interfaces()]
        )

    @staticmethod
    def wait_for_network() -> None:
        """Wait for network to come up."""
        raise NotImplementedError()


class IfUpDownActivator(NetworkActivator):
    # Note that we're not overriding bring_up_interfaces to pass something
    # like ifup --all because it isn't supported everywhere.
    # E.g., NetworkManager has a ifupdown plugin that requires the name
    # of a specific connection.
    @staticmethod
    def available(target: Optional[str] = None) -> bool:
        """Return true if ifupdown can be used on this system."""
        return eni.available(target=target)

    @staticmethod
    def bring_up_interface(device_name: str) -> bool:
        """Bring up interface using ifup.

        Return True is successful, otherwise return False
        """
        cmd = ["ifup", device_name]
        return _alter_interface(cmd, device_name)

    @staticmethod
    def bring_down_interface(device_name: str) -> bool:
        """Bring up interface using ifup.

        Return True is successful, otherwise return False
        """
        cmd = ["ifdown", device_name]
        return _alter_interface(cmd, device_name)


class IfConfigActivator(NetworkActivator):
    @staticmethod
    def available(target=None) -> bool:
        """Return true if ifconfig can be used on this system."""
        expected = "ifconfig"
        search = ["/sbin"]
        return bool(subp.which(expected, search=search, target=target))

    @staticmethod
    def bring_up_interface(device_name: str) -> bool:
        """Bring up interface using ifconfig <dev> up.

        Return True is successful, otherwise return False
        """
        cmd = ["ifconfig", device_name, "up"]
        return _alter_interface(cmd, device_name)

    @staticmethod
    def bring_down_interface(device_name: str) -> bool:
        """Bring up interface using ifconfig <dev> down.

        Return True is successful, otherwise return False
        """
        cmd = ["ifconfig", device_name, "down"]
        return _alter_interface(cmd, device_name)


class NetworkManagerActivator(NetworkActivator):
    @staticmethod
    def available(target=None) -> bool:
        """Return true if NetworkManager can be used on this system."""
        return network_manager.available(target=target)

    @staticmethod
    def bring_up_interface(device_name: str) -> bool:
        """Bring up connection using nmcli.

        Return True is successful, otherwise return False
        """
        from cloudinit.net.network_manager import conn_filename

        filename = conn_filename(device_name)
        if filename is None:
            LOG.warning(
                "Unable to find an interface config file. "
                "Unable to bring up interface."
            )
            return False

        cmd = ["nmcli", "connection", "load", filename]
        if _alter_interface(cmd, device_name):
            cmd = ["nmcli", "connection", "up", "filename", filename]
        else:
            _alter_interface(["nmcli", "connection", "reload"], device_name)
            cmd = ["nmcli", "connection", "up", "ifname", device_name]
        return _alter_interface(cmd, device_name)

    @staticmethod
    def bring_down_interface(device_name: str) -> bool:
        """Bring down interface using nmcli.

        Return True is successful, otherwise return False
        """
        cmd = ["nmcli", "device", "disconnect", device_name]
        return _alter_interface(cmd, device_name)

    @classmethod
    def bring_up_interfaces(cls, device_names: Iterable[str]) -> bool:
        """Activate network

        Return True on success
        """
        state = subp.subp(
            [
                "systemctl",
                "show",
                "--property=SubState",
                "NetworkManager.service",
            ]
        ).stdout.rstrip()
        if "SubState=running" != state:
            LOG.warning(
                "Expected NetworkManager SubState=running, but detected: %s",
                state,
            )
        return _alter_interface(
            ["systemctl", "try-reload-or-restart", "NetworkManager.service"],
            "all",
        ) and all(cls.bring_up_interface(device) for device in device_names)


class NetplanActivator(NetworkActivator):
    NETPLAN_CMD = ["netplan", "apply"]

    @staticmethod
    def available(target=None) -> bool:
        """Return true if netplan can be used on this system."""
        return netplan.available(target=target)

    @staticmethod
    def bring_up_interface(device_name: str) -> bool:
        """Apply netplan config.

        Return True is successful, otherwise return False
        """
        LOG.debug(
            "Calling 'netplan apply' rather than "
            "altering individual interfaces"
        )
        return _alter_interface(
            NetplanActivator.NETPLAN_CMD, "all", warn_on_stderr=False
        )

    @staticmethod
    def bring_up_interfaces(device_names: Iterable[str]) -> bool:
        """Apply netplan config.

        Return True is successful, otherwise return False
        """
        LOG.debug(
            "Calling 'netplan apply' rather than "
            "altering individual interfaces"
        )
        return _alter_interface(
            NetplanActivator.NETPLAN_CMD, "all", warn_on_stderr=False
        )

    @staticmethod
    def bring_up_all_interfaces(network_state: NetworkState) -> bool:
        """Apply netplan config.

        Return True is successful, otherwise return False
        """
        return _alter_interface(
            NetplanActivator.NETPLAN_CMD, "all", warn_on_stderr=False
        )

    @staticmethod
    def bring_down_interface(device_name: str) -> bool:
        """Apply netplan config.

        Return True is successful, otherwise return False
        """
        LOG.debug(
            "Calling 'netplan apply' rather than "
            "altering individual interfaces"
        )
        return _alter_interface(
            NetplanActivator.NETPLAN_CMD, "all", warn_on_stderr=False
        )

    @staticmethod
    def wait_for_network() -> None:
        """On networkd systems, wait for systemd-networkd-wait-online"""
        # At the moment, this is only supported using the networkd renderer.
        if network_manager.available():
            LOG.debug("NetworkManager is enabled, skipping networkd wait")
            return
        NetworkdActivator.wait_for_network()


class NetworkdActivator(NetworkActivator):
    @staticmethod
    def available(target=None) -> bool:
        """Return true if ifupdown can be used on this system."""
        return networkd.available(target=target)

    @staticmethod
    def bring_up_interface(device_name: str) -> bool:
        """Return True is successful, otherwise return False"""
        return _alter_interface_callable(
            partial(Iproute2.link_up, device_name)
        )

    @staticmethod
    def bring_up_all_interfaces(network_state: NetworkState) -> bool:
        """Return True is successful, otherwise return False"""
        cmd = ["systemctl", "restart", "systemd-networkd", "systemd-resolved"]
        return _alter_interface(cmd, "all")

    @staticmethod
    def bring_down_interface(device_name: str) -> bool:
        """Return True is successful, otherwise return False"""
        return _alter_interface_callable(
            partial(Iproute2.link_down, device_name)
        )

    @staticmethod
    def wait_for_network() -> None:
        """Wait for systemd-networkd-wait-online."""
        subp.subp(
            ["systemctl", "start", "systemd-networkd-wait-online.service"]
        )


# This section is mostly copied and pasted from renderers.py. An abstract
# version to encompass both seems overkill at this point
DEFAULT_PRIORITY = [
    "eni",
    "netplan",
    "network-manager",
    "networkd",
    "ifconfig",
]

NAME_TO_ACTIVATOR: Dict[str, Type[NetworkActivator]] = {
    "eni": IfUpDownActivator,
    "netplan": NetplanActivator,
    "network-manager": NetworkManagerActivator,
    "networkd": NetworkdActivator,
    "ifconfig": IfConfigActivator,
}


def search_activator(
    priority: List[str], target: Union[str, None]
) -> Optional[Type[NetworkActivator]]:
    """Returns the first available activator from the priority list or None."""
    unknown = [i for i in priority if i not in DEFAULT_PRIORITY]
    if unknown:
        raise ValueError(
            f"Unknown activators provided in priority list: {unknown}"
        )
    activator_classes = [NAME_TO_ACTIVATOR[name] for name in priority]
    return next(
        (
            activator_cls
            for activator_cls in activator_classes
            if activator_cls.available(target)
        ),
        None,
    )


def select_activator(
    priority: Optional[List[str]] = None, target: Optional[str] = None
) -> Type[NetworkActivator]:
    if priority is None:
        priority = DEFAULT_PRIORITY
    selected = search_activator(priority, target)
    if not selected:
        tmsg = f" in target={target}" if target and target != "/" else ""
        raise NoActivatorException(
            f"No available network activators found{tmsg}. "
            f"Searched through list: {priority}"
        )
    LOG.debug(
        "Using selected activator: %s from priority: %s", selected, priority
    )
    return selected



# This file is part of cloud-init. See LICENSE file for license information.

import base64
import gzip
import io
import logging
import re
import shutil
import tempfile

import pytest
import responses

from cloudinit import util
from cloudinit.config.cc_write_files import decode_perms, handle, write_files
from cloudinit.config.schema import (
    SchemaValidationError,
    get_schema,
    validate_cloudconfig_schema,
)
from tests.unittests.helpers import (
    SCHEMA_EMPTY_ERROR,
    CiTestCase,
    FilesystemMockingTestCase,
    skipUnlessJsonSchema,
)

LOG = logging.getLogger(__name__)

YAML_TEXT = """
write_files:
 - encoding: gzip
   content: !!binary |
     H4sIAIDb/U8C/1NW1E/KzNMvzuBKTc7IV8hIzcnJVyjPL8pJ4QIA6N+MVxsAAAA=
   path: /usr/bin/hello
   permissions: '0755'
 - content: !!binary |
     Zm9vYmFyCg==
   path: /wark
   permissions: '0755'
 - content: |
    hi mom line 1
    hi mom line 2
   path: /tmp/message
"""

YAML_CONTENT_EXPECTED = {
    "/usr/bin/hello": "#!/bin/sh\necho hello world\n",
    "/wark": "foobar\n",
    "/tmp/message": "hi mom line 1\nhi mom line 2\n",
}

VALID_SCHEMA = {
    "write_files": [
        {
            "append": False,
            "content": "a",
            "encoding": "gzip",
            "owner": "jeff",
            "path": "/some",
            "permissions": "0777",
        }
    ]
}


class TestWriteFiles(FilesystemMockingTestCase):

    with_logs = True
    owner = "root:root"

    def setUp(self):
        super(TestWriteFiles, self).setUp()
        self.tmp = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.tmp)

    def test_simple(self):
        self.patchUtils(self.tmp)
        expected = "hello world\n"
        filename = "/tmp/my.file"
        write_files(
            "test_simple",
            [{"content": expected, "path": filename}],
            self.owner,
        )
        self.assertEqual(util.load_text_file(filename), expected)

    def test_empty(self):
        self.patchUtils(self.tmp)
        filename = "/tmp/my.file"
        write_files(
            "test_empty",
            [{"path": filename}],
            self.owner,
        )
        self.assertEqual(util.load_text_file(filename), "")

    def test_append(self):
        self.patchUtils(self.tmp)
        existing = "hello "
        added = "world\n"
        expected = existing + added
        filename = "/tmp/append.file"
        util.write_file(filename, existing)
        write_files(
            "test_append",
            [{"content": added, "path": filename, "append": "true"}],
            self.owner,
        )
        self.assertEqual(util.load_text_file(filename), expected)

    def test_yaml_binary(self):
        self.patchUtils(self.tmp)
        data = util.load_yaml(YAML_TEXT)
        write_files("testname", data["write_files"], self.owner)
        for path, content in YAML_CONTENT_EXPECTED.items():
            self.assertEqual(util.load_text_file(path), content)

    def test_all_decodings(self):
        self.patchUtils(self.tmp)

        # build a 'files' array that has a dictionary of encodings
        # for 'gz', 'gzip', 'gz+base64' ...
        data = b"foobzr"
        utf8_valid = b"foobzr"
        utf8_invalid = b"ab\xaadef"
        files = []
        expected = []

        gz_aliases = ("gz", "gzip")
        gz_b64_aliases = ("gz+base64", "gzip+base64", "gz+b64", "gzip+b64")
        b64_aliases = ("base64", "b64")

        datum = (("utf8", utf8_valid), ("no-utf8", utf8_invalid))
        for name, data in datum:
            gz = (_gzip_bytes(data), gz_aliases)
            gz_b64 = (base64.b64encode(_gzip_bytes(data)), gz_b64_aliases)
            b64 = (base64.b64encode(data), b64_aliases)
            for content, aliases in (gz, gz_b64, b64):
                for enc in aliases:
                    path = "/tmp/file-%s-%s" % (name, enc)
                    cur = {
                        "content": content,
                        "path": path,
                        "encoding": enc,
                    }
                    files.append(cur)
                    expected.append((path, data))

        write_files("test_decoding", files, self.owner)

        for path, content in expected:
            self.assertEqual(util.load_binary_file(path), content)

        # make sure we actually wrote *some* files.
        flen_expected = len(gz_aliases + gz_b64_aliases + b64_aliases) * len(
            datum
        )
        self.assertEqual(len(expected), flen_expected)

    def test_handle_plain_text(self):
        self.patchUtils(self.tmp)
        file_path = "/tmp/file-text-plain"
        content = "asdf"
        cfg = {
            "write_files": [
                {
                    "content": content,
                    "path": file_path,
                    "encoding": "text/plain",
                    "defer": False,
                }
            ]
        }
        cc = self.tmp_cloud("ubuntu")
        handle("ignored", cfg, cc, [])
        assert content == util.load_text_file(file_path)
        self.assertNotIn(
            "Unknown encoding type text/plain", self.logs.getvalue()
        )

    def test_file_uri(self):
        self.patchUtils(self.tmp)
        src_path = "/tmp/file-uri"
        dst_path = "/tmp/file-uri-target"
        content = "asdf"
        util.write_file(src_path, content)
        cfg = {
            "write_files": [
                {
                    "source": {"uri": "file://" + src_path},
                    "path": dst_path,
                }
            ]
        }
        cc = self.tmp_cloud("ubuntu")
        handle("ignored", cfg, cc, [])
        self.assertEqual(
            util.load_text_file(src_path), util.load_text_file(dst_path)
        )

    @responses.activate
    def test_http_uri(self):
        self.patchUtils(self.tmp)
        path = "/tmp/http-uri-target"
        url = "http://hostname/path"
        content = "more asdf"
        responses.add(responses.GET, url, content)
        cfg = {
            "write_files": [
                {
                    "source": {
                        "uri": url,
                        "headers": {
                            "foo": "bar",
                            "blah": "blah",
                        },
                    },
                    "path": path,
                }
            ]
        }
        cc = self.tmp_cloud("ubuntu")
        handle("ignored", cfg, cc, [])
        self.assertEqual(content, util.load_text_file(path))

    def test_uri_fallback(self):
        self.patchUtils(self.tmp)
        src_path = "/tmp/INVALID"
        dst_path = "/tmp/uri-fallback-target"
        content = "asdf"
        util.del_file(src_path)
        cfg = {
            "write_files": [
                {
                    "source": {"uri": "file://" + src_path},
                    "content": content,
                    "encoding": "text/plain",
                    "path": dst_path,
                }
            ]
        }
        cc = self.tmp_cloud("ubuntu")
        handle("ignored", cfg, cc, [])
        self.assertEqual(content, util.load_text_file(dst_path))

    def test_deferred(self):
        self.patchUtils(self.tmp)
        file_path = "/tmp/deferred.file"
        config = {"write_files": [{"path": file_path, "defer": True}]}
        cc = self.tmp_cloud("ubuntu")
        handle("cc_write_file", config, cc, [])
        with self.assertRaises(FileNotFoundError):
            util.load_text_file(file_path)


class TestDecodePerms(CiTestCase):

    with_logs = True

    def test_none_returns_default(self):
        """If None is passed as perms, then default should be returned."""
        default = object()
        found = decode_perms(None, default)
        self.assertEqual(default, found)

    def test_integer(self):
        """A valid integer should return itself."""
        found = decode_perms(0o755, None)
        self.assertEqual(0o755, found)

    def test_valid_octal_string(self):
        """A string should be read as octal."""
        found = decode_perms("644", None)
        self.assertEqual(0o644, found)

    def test_invalid_octal_string_returns_default_and_warns(self):
        """A string with invalid octal should warn and return default."""
        found = decode_perms("999", None)
        self.assertIsNone(found)
        self.assertIn("WARNING: Undecodable", self.logs.getvalue())


def _gzip_bytes(data):
    buf = io.BytesIO()
    fp = None
    try:
        fp = gzip.GzipFile(fileobj=buf, mode="wb")
        fp.write(data)
        fp.close()
        return buf.getvalue()
    finally:
        if fp:
            fp.close()


class TestWriteFilesSchema:
    @pytest.mark.parametrize(
        "config, error_msg",
        [
            # Top-level write_files type validation
            ({"write_files": 1}, "write_files: 1 is not of type 'array'"),
            (
                {"write_files": []},
                re.escape("write_files: [] ") + SCHEMA_EMPTY_ERROR,
            ),
            (
                {"write_files": [{}]},
                "write_files.0: 'path' is a required property",
            ),
            (
                {"write_files": [{"path": "/some", "bogus": True}]},
                re.escape(
                    "write_files.0: Additional properties are not allowed"
                    " ('bogus'"
                ),
            ),
            (  # Strict encoding choices
                {"write_files": [{"path": "/some", "encoding": "g"}]},
                re.escape(
                    "write_files.0.encoding: 'g' is not one of ['gz', 'gzip',"
                ),
            ),
            (
                {
                    "write_files": [
                        {
                            "append": False,
                            "source": {
                                "uri": "http://a.com/a",
                                "headers": {
                                    "Authorization": "Bearer SOME_TOKEN"
                                },
                            },
                            "content": "a",
                            "encoding": "text/plain",
                            "owner": "jeff",
                            "path": "/some",
                            "permissions": "0777",
                        }
                    ]
                },
                None,
            ),
        ],
    )
    @skipUnlessJsonSchema()
    def test_schema_validation(self, config, error_msg):
        if error_msg is not None:
            with pytest.raises(SchemaValidationError, match=error_msg):
                validate_cloudconfig_schema(config, get_schema(), strict=True)
        else:
            validate_cloudconfig_schema(config, get_schema(), strict=True)



"""cloudinit.signal_handler tests"""

import inspect
import signal
from unittest.mock import Mock, patch

import pytest

from cloudinit import signal_handler

REENTRANT = "reentrant"


class TestSignalHandler:

    @pytest.mark.parametrize(
        "m_args",
        [
            (signal.SIGINT, inspect.currentframe()),
            (9, None),
            (signal.SIGTERM, None),
            (1, inspect.currentframe()),
        ],
    )
    @pytest.mark.parametrize(
        "m_suspended",
        [
            (REENTRANT, 0),
            (True, 0),
            (False, 1),
        ],
    )
    def test_suspend_signal(self, m_args, m_suspended):
        """suspend_crash should prevent crashing (exit 1) on signal

        otherwise cloud-init should exit 1
        """
        sig, frame = m_args
        suspended, rc = m_suspended

        with patch.object(signal_handler.sys, "exit", Mock()) as m_exit:
            if suspended is True:
                with signal_handler.suspend_crash():
                    signal_handler._handle_exit(sig, frame)
            elif suspended == REENTRANT:
                with signal_handler.suspend_crash():
                    with signal_handler.suspend_crash():
                        signal_handler._handle_exit(sig, frame)
            else:
                signal_handler._handle_exit(sig, frame)
        m_exit.assert_called_with(rc)



"""reaper self-test"""

import logging
import time
import warnings
from unittest import mock

import pytest

from tests.integration_tests import reaper
from tests.integration_tests.instances import IntegrationInstance

LOG = logging.Logger(__name__)


class MockInstance(IntegrationInstance):
    # because of instance id printing
    instance = mock.Mock()

    def __init__(self, times_refused):
        self.times_refused = times_refused
        self.call_count = 0

        # assert that destruction succeeded
        self.stopped = False

    def destroy(self):
        """destroy() only succeeds after failing N=times_refused times"""
        if self.call_count == self.times_refused:
            self.stopped = True
            return
        self.call_count += 1
        raise RuntimeError("I object!")


@pytest.mark.ci
class TestReaper:
    def test_start_stop(self):
        """basic setup teardown"""

        instance = MockInstance(0)
        r = reaper._Reaper()
        # start / stop
        r.start()
        r.stop()
        # start / reap / stop
        r.start()
        r.reap(instance)
        r.stop()

        # start / stop
        r.start()
        r.stop()
        assert instance.stopped

    def test_basic_reap(self):
        """basic setup teardown"""

        i_1 = MockInstance(0)
        r = reaper._Reaper()
        r.start()
        r.reap(i_1)
        r.stop()
        assert i_1.stopped

    def test_unreaped_instance(self):
        """a single warning should print for any number of leaked instances"""

        i_1 = MockInstance(64)
        i_2 = MockInstance(64)
        r = reaper._Reaper()
        r.start()
        r.reap(i_1)
        r.reap(i_2)
        with warnings.catch_warnings(record=True) as w:
            r.stop()
        assert len(w) == 1

    def test_stubborn_reap(self):
        """verify that stubborn instances are cleaned"""

        sleep_time = 0.000_001
        sleep_total = 0.0
        instances = [
            MockInstance(0),
            MockInstance(3),
            MockInstance(6),
            MockInstance(9),
            MockInstance(12),
            MockInstance(9),
            MockInstance(6),
            MockInstance(3),
            MockInstance(0),
        ]

        # forcibly disallow sleeping, to avoid wasted time during tests
        r = reaper._Reaper(timeout=0.0)
        r.start()
        for i in instances:
            r.reap(i)

        # this should really take no time at all, waiting 1s should be plenty
        # of time for the reaper to reap it when not sleeping
        while sleep_total < 1.0:
            # are any still undead?
            any_undead = False
            for i in instances:
                if not i.stopped:
                    any_undead = True
                    break
            if not any_undead:
                # test passed
                # Advance to GO, collect $400
                break
            # sleep then recheck, incremental backoff
            sleep_total += sleep_time
            sleep_time *= 2
            time.sleep(sleep_time)
        r.stop()
        for i in instances:
            assert i.stopped, (
                f"Reaper didn't reap stubborn instance {i} in {sleep_total}s. "
                "Something appears to be broken in the reaper logic or test."
            )

    def test_start_stop_multiple(self):
        """reap lots of instances

        obedient ones
        """
        num = 64
        instances = []
        r = reaper._Reaper()
        r.start()
        for _ in range(num):
            i = MockInstance(0)
            instances.append(i)
            r.reap(i)
        r.stop()
        for i in instances:
            assert i.stopped



# This file is part of cloud-init. See LICENSE file for license information.

import glob
import os
import pathlib
import sys
import tarfile
from datetime import datetime, timezone

import pytest

from cloudinit.cmd.devel import logs
from cloudinit.cmd.devel.logs import ApportFile
from cloudinit.subp import SubpResult, subp
from cloudinit.util import ensure_dir, load_text_file, write_file

M_PATH = "cloudinit.cmd.devel.logs."
INSTANCE_JSON_SENSITIVE_FILE = "instance-data-sensitive.json"


def fake_subp(cmd):
    if cmd[0] == "tar" and cmd[1] == "czf":
        subp(cmd)  # Pass through tar cmd so we can check output
        return SubpResult("", "")

    expected_subp = {
        (
            "dpkg-query",
            "--show",
            "-f=${Version}\n",
            "cloud-init",
        ): "0.7fake\n",
        ("cloud-init", "--version"): "over 9000\n",
    }
    cmd_tuple = tuple(cmd)
    if cmd_tuple not in expected_subp:
        raise AssertionError(
            "Unexpected command provided to subp: {0}".format(cmd)
        )

    return SubpResult(expected_subp[cmd_tuple], "")


# the new _stream_command_output_to_file function uses subprocess.call
# instead of subp, so we need to mock that as well
def fake_subprocess_call(cmd, stdout=None, stderr=None):
    expected_calls = {
        ("dmesg",): "dmesg-out\n",
        ("journalctl", "--boot=0", "-o", "short-precise"): "journal-out\n",
        ("journalctl", "--boot=-1", "-o", "short-precise"): "journal-prev\n",
    }
    cmd_tuple = tuple(cmd)
    if cmd_tuple not in expected_calls:
        raise AssertionError(
            "Unexpected command provided to subprocess: {0}".format(cmd)
        )
    stdout.write(expected_calls[cmd_tuple])


def patch_subiquity_paths(mocker, tmp_path):
    mocker.patch(
        "cloudinit.cmd.devel.logs.INSTALLER_APPORT_FILES",
        [
            ApportFile(
                str(tmp_path / "subiquity-server-debug.log"),
                "subiquityServerDebug",
            )
        ],
    )
    mocker.patch(
        "cloudinit.cmd.devel.logs.INSTALLER_APPORT_SENSITIVE_FILES",
        [
            ApportFile(
                str(tmp_path / "autoinstall-user-data"), "AutoInstallUserData"
            )
        ],
    )


class TestCollectLogs:
    def test_collect_logs_requires_root_user(self, mocker):
        """collect-logs errors when non-root user collects userdata ."""
        # 100 is non-root
        mocker.patch("cloudinit.cmd.devel.logs.os.getuid", retrn_value=100)
        # If we don't mock this, we can change logging for future tests
        mocker.patch("cloudinit.cmd.devel.logs._setup_logger")
        with pytest.raises(
            RuntimeError, match="This command must be run as root"
        ):
            logs.collect_logs_cli("")

    def test_collect_logs_end_to_end(self, mocker, tmp_path):
        mocker.patch(f"{M_PATH}subp", side_effect=fake_subp)
        mocker.patch(
            f"{M_PATH}subprocess.call", side_effect=fake_subprocess_call
        )
        mocker.patch(
            f"{M_PATH}_get_etc_cloud",
            return_value=[
                tmp_path / "etc/cloud/cloud.cfg",
                tmp_path / "etc/cloud/cloud.cfg.d/90-dpkg.cfg",
            ],
        )
        patch_subiquity_paths(mocker, tmp_path)
        today = datetime.now(timezone.utc).date().strftime("%Y-%m-%d")

        # This list isn't exhaustive
        to_collect = [
            "etc/cloud/cloud.cfg",
            "etc/cloud/cloud.cfg.d/90-dpkg.cfg",
            "var/lib/cloud/instance/instance-id",
            "var/lib/cloud/instance/user-data.txt",
            "var/lib/cloud/instance/user-data.txt.i",
            "var/lib/cloud/handlers/wtf-i-wrote-a-handler.py",
            "var/log/cloud-init.log",
            "var/log/cloud-init-output.log",
            "var/log/cloud-init.log.1.gz",
            "var/log/cloud-init-output.log.1.gz",
            "run/cloud-init/results.json",
            "run/cloud-init/status.json",
            "run/cloud-init/instance-data-sensitive.json",
            "run/cloud-init/instance-data.json",
            "subiquity-server-debug.log",
            "autoinstall-user-data",
        ]
        for to_write in to_collect:
            write_file(tmp_path / to_write, pathlib.Path(to_write).name)

        # logs.collect_logs("cloud-init.tar.gz", {})
        logs.collect_logs(
            tarfile=tmp_path / "cloud-init.tar.gz",
            log_cfg={
                "def_log_file": str(tmp_path / "var/log/cloud-init.log"),
                "output": {
                    "all": f"| tee -a {tmp_path}/var/log/cloud-init-output.log"
                },
            },
            run_dir=tmp_path / "run/cloud-init",
            cloud_dir=tmp_path / "var/lib/cloud",
            include_sensitive=True,
        )
        extract_to = tmp_path / "extracted"
        extract_to.mkdir()

        tar_kwargs = {}
        if sys.version_info > (3, 11):
            tar_kwargs = {"filter": "fully_trusted"}
        with tarfile.open(tmp_path / "cloud-init.tar.gz") as tar:
            tar.extractall(extract_to, **tar_kwargs)  # type: ignore[arg-type]
        extracted_dir = extract_to / f"cloud-init-logs-{today}"

        for name in to_collect:
            # Since we've collected absolute paths, that means even though
            # our extract contents are within the tmp_path, the files will
            # include another layer of tmp_path directories
            assert (extracted_dir / str(tmp_path)[1:] / name).exists()

        assert (extracted_dir / "journal.txt").read_text() == "journal-out\n"
        assert (extracted_dir / "dmesg.txt").read_text() == "dmesg-out\n"
        assert (extracted_dir / "dpkg-version").read_text() == "0.7fake\n"
        assert (extracted_dir / "version").read_text() == "over 9000\n"

    def test_logs_and_installer_ignore_sensitive_flag(self, mocker, tmp_path):
        """Regardless of the sensitive flag, we always want these logs."""
        mocker.patch(f"{M_PATH}subp", side_effect=fake_subp)
        mocker.patch(
            f"{M_PATH}subprocess.call", side_effect=fake_subprocess_call
        )
        mocker.patch(f"{M_PATH}_get_etc_cloud", return_value=[])
        patch_subiquity_paths(mocker, tmp_path)

        to_collect = [
            "var/log/cloud-init.log",
            "var/log/cloud-init-output.log",
            "var/log/cloud-init.log.1.gz",
            "var/log/cloud-init-output.log.1.gz",
            "subiquity-server-debug.log",
        ]

        for to_write in to_collect:
            write_file(
                tmp_path / to_write, pathlib.Path(to_write).name, mode=0x700
            )

        collect_dir = tmp_path / "collect"
        collect_dir.mkdir()
        logs._collect_logs_into_tmp_dir(
            log_dir=collect_dir,
            log_cfg={
                "def_log_file": str(tmp_path / "var/log/cloud-init.log"),
                "output": {
                    "all": f"| tee -a {tmp_path}/var/log/cloud-init-output.log"
                },
            },
            run_dir=collect_dir,
            cloud_dir=collect_dir,
            include_sensitive=False,
        )

        for name in to_collect:
            assert (collect_dir / str(tmp_path)[1:] / name).exists()

    def test_root_read_only_not_collected_on_redact(self, mocker, tmp_path):
        """Don't collect root read-only files."""
        mocker.patch(f"{M_PATH}subp", side_effect=fake_subp)
        mocker.patch(
            f"{M_PATH}subprocess.call", side_effect=fake_subprocess_call
        )
        mocker.patch(f"{M_PATH}_get_etc_cloud", return_value=[])
        patch_subiquity_paths(mocker, tmp_path)

        to_collect = [
            "etc/cloud/cloud.cfg",
            "etc/cloud/cloud.cfg.d/90-dpkg.cfg",
            "var/lib/cloud/instance/instance-id",
            "var/lib/cloud/instance/user-data.txt",
            "var/lib/cloud/instance/user-data.txt.i",
            "var/lib/cloud/handlers/wtf-i-wrote-a-handler.py",
            "run/cloud-init/results.json",
            "run/cloud-init/status.json",
            "run/cloud-init/instance-data-sensitive.json",
            "run/cloud-init/instance-data.json",
            "autoinstall-user-data",
        ]

        for to_write in to_collect:
            write_file(
                tmp_path / to_write, pathlib.Path(to_write).name, mode=0x700
            )

        collect_dir = tmp_path / "collect"
        collect_dir.mkdir()
        logs._collect_logs_into_tmp_dir(
            log_dir=collect_dir,
            log_cfg={
                "def_log_file": str(tmp_path / "var/log/cloud-init.log"),
                "output": {
                    "all": f"| tee -a {tmp_path}/var/log/cloud-init-output.log"
                },
            },
            run_dir=collect_dir,
            cloud_dir=collect_dir,
            include_sensitive=False,
        )

        for name in to_collect:
            assert not (collect_dir / str(tmp_path)[1:] / name).exists()
        assert not (collect_dir / "dmsg.txt").exists()

    @pytest.mark.parametrize(
        "cmd, expected_file_contents, expected_return_value",
        [
            (
                ["echo", "cloud-init? more like cloud-innit!"],
                "cloud-init? more like cloud-innit!\n",
                "cloud-init? more like cloud-innit!\n",
            ),
            (
                ["sh", "-c", "echo test 1>&2; exit 42"],
                (
                    "Unexpected error while running command.\n"
                    "Command: ['sh', '-c', 'echo test 1>&2; exit 42']\n"
                    "Exit code: 42\n"
                    "Reason: -\n"
                    "Stdout: \n"
                    "Stderr: test"
                ),
                None,
            ),
        ],
    )
    def test_write_command_output_to_file(
        self,
        tmp_path,
        cmd,
        expected_file_contents,
        expected_return_value,
    ):
        output_file = tmp_path / "test-output-file.txt"

        return_output = logs._write_command_output_to_file(
            cmd=cmd,
            file_path=output_file,
            msg="",
        )

        assert expected_return_value == return_output
        assert expected_file_contents == load_text_file(output_file)

    @pytest.mark.parametrize(
        "cmd, expected_file_contents",
        [
            (["echo", "cloud-init, shmoud-init"], "cloud-init, shmoud-init\n"),
            (["sh", "-c", "echo test 1>&2; exit 42"], "test\n"),
        ],
    )
    def test_stream_command_output_to_file(
        self, tmp_path, cmd, expected_file_contents
    ):
        output_file = tmp_path / "test-output-file.txt"

        logs._stream_command_output_to_file(
            cmd=cmd,
            file_path=output_file,
            msg="",
        )

        assert expected_file_contents == load_text_file(output_file)


class TestCollectInstallerLogs:
    @pytest.mark.parametrize(
        "include_sensitive, apport_files, apport_sensitive_files",
        (
            pytest.param(True, [], [], id="no_files_include_userdata"),
            pytest.param(False, [], [], id="no_files_exclude_userdata"),
            pytest.param(
                True,
                (ApportFile("log1", "Label1"), ApportFile("log2", "Label2")),
                (
                    ApportFile("private1", "LabelPrivate1"),
                    ApportFile("private2", "PrivateLabel2"),
                ),
                id="files_and_dirs_include_userdata",
            ),
            pytest.param(
                False,
                (ApportFile("log1", "Label1"), ApportFile("log2", "Label2")),
                (
                    ApportFile("private1", "LabelPrivate1"),
                    ApportFile("private2", "PrivateLabel2"),
                ),
                id="files_and_dirs_exclude_userdata",
            ),
        ),
    )
    def test_include_installer_logs_when_present(
        self,
        include_sensitive,
        apport_files,
        apport_sensitive_files,
        tmpdir,
        mocker,
    ):
        src_dir = tmpdir.join("src")
        ensure_dir(src_dir.strpath)
        # collect-logs nests full directory path to file in the tarfile
        destination_dir = tmpdir.join(src_dir)

        # Create tmppath-based userdata_files, installer_logs, installer_dirs
        expected_files = []
        # Create last file in list to assert ignoring absent files
        apport_files = [
            logs.ApportFile(src_dir.join(apport.path).strpath, apport.label)
            for apport in apport_files
        ]
        if apport_files:
            write_file(apport_files[-1].path, apport_files[-1].label)
            expected_files += [
                destination_dir.join(
                    os.path.basename(apport_files[-1].path)
                ).strpath
            ]
        apport_sensitive_files = [
            logs.ApportFile(src_dir.join(apport.path).strpath, apport.label)
            for apport in apport_sensitive_files
        ]
        if apport_sensitive_files:
            write_file(
                apport_sensitive_files[-1].path,
                apport_sensitive_files[-1].label,
            )
            if include_sensitive:
                expected_files += [
                    destination_dir.join(
                        os.path.basename(apport_sensitive_files[-1].path)
                    ).strpath
                ]
        mocker.patch(M_PATH + "INSTALLER_APPORT_FILES", apport_files)
        mocker.patch(
            M_PATH + "INSTALLER_APPORT_SENSITIVE_FILES", apport_sensitive_files
        )
        logs._collect_installer_logs(
            log_dir=tmpdir.strpath,
            include_sensitive=include_sensitive,
        )
        expect_userdata = bool(include_sensitive and apport_sensitive_files)
        # when subiquity artifacts exist, and userdata set true, expect logs
        expect_subiquity_logs = any([apport_files, expect_userdata])
        if expect_subiquity_logs:
            assert destination_dir.exists(), "Missing subiquity artifact dir"
            assert sorted(expected_files) == sorted(
                glob.glob(f"{destination_dir.strpath}/*")
            )
        else:
            assert not destination_dir.exists(), "Unexpected subiquity dir"



import logging

import pytest

from cloudinit import lifecycle

LOG = logging.getLogger()


class TestLogWithDowngradableLevel:
    @pytest.mark.parametrize(
        "version,expected",
        [
            ("9", logging.ERROR),
            ("11", logging.DEBUG),
        ],
    )
    def test_log_with_downgradable_level(
        self, mocker, caplog, version, expected
    ):
        mocker.patch("cloudinit.features.DEPRECATION_INFO_BOUNDARY", "10")
        lifecycle.log_with_downgradable_level(
            logger=LOG,
            version=version,
            requested_level=logging.ERROR,
            msg="look at me %s %s!",
            args=("one", "two"),
        )
        records = caplog.record_tuples
        assert len(records) == 1
        assert records[0][1] == expected
        assert records[0][2] == "look at me one two!"



# This file is part of cloud-init. See LICENSE file for license information.

import copy
import logging
from typing import List, Union

from cloudinit import dmi, sources
from cloudinit import url_helper as uhelp
from cloudinit import util
from cloudinit.event import EventScope, EventType
from cloudinit.net.dhcp import NoDHCPLeaseError
from cloudinit.net.ephemeral import EphemeralIPNetwork
from cloudinit.sources import DataSourceHostname
from cloudinit.sources.helpers import aliyun, ec2

LOG = logging.getLogger(__name__)

ALIYUN_PRODUCT = "Alibaba Cloud ECS"


class DataSourceAliYun(sources.DataSource):

    dsname = "AliYun"
    metadata_urls = ["http://100.100.100.200"]

    # The minimum supported metadata_version from the ecs metadata apis
    min_metadata_version = "2016-01-01"
    extended_metadata_versions: List[str] = []

    # Setup read_url parameters per get_url_params.
    url_max_wait = 240
    url_timeout = 50

    # API token for accessing the metadata service
    _api_token = None
    # Used to cache calculated network cfg v1
    _network_config: Union[str, dict] = sources.UNSET

    # Whether we want to get network configuration from the metadata service.
    perform_dhcp_setup = False

    # Aliyun metadata server security enhanced mode overwrite
    @property
    def imdsv2_token_put_header(self):
        return "X-aliyun-ecs-metadata-token"

    def __init__(self, sys_cfg, distro, paths):
        super(DataSourceAliYun, self).__init__(sys_cfg, distro, paths)
        self.default_update_events = copy.deepcopy(self.default_update_events)
        self.default_update_events[EventScope.NETWORK].add(EventType.BOOT)

    def _unpickle(self, ci_pkl_version: int) -> None:
        super()._unpickle(ci_pkl_version)

    def get_hostname(self, fqdn=False, resolve_ip=False, metadata_only=False):
        hostname = self.metadata.get("hostname")
        is_default = False
        if hostname is None:
            hostname = "localhost.localdomain"
            is_default = True
        return DataSourceHostname(hostname, is_default)

    def get_public_ssh_keys(self):
        return parse_public_keys(self.metadata.get("public-keys", {}))

    def _get_cloud_name(self):
        if _is_aliyun():
            return self.dsname.lower()
        return "NO_ALIYUN_METADATA"

    @property
    def platform(self):
        return self.dsname.lower()

    # IMDSv2 related parameters from the ecs metadata api document
    @property
    def api_token_route(self):
        return "latest/api/token"

    @property
    def imdsv2_token_ttl_seconds(self):
        return "21600"

    @property
    def imdsv2_token_redact(self):
        return [self.imdsv2_token_put_header, self.imdsv2_token_req_header]

    @property
    def imdsv2_token_req_header(self):
        return self.imdsv2_token_put_header + "-ttl-seconds"

    @property
    def network_config(self):
        """Return a network config dict for rendering ENI or netplan files."""
        if self._network_config != sources.UNSET:
            return self._network_config

        result = {}
        iface = self.distro.fallback_interface
        net_md = self.metadata.get("network")
        if isinstance(net_md, dict):
            result = aliyun.convert_ecs_metadata_network_config(
                net_md,
                fallback_nic=iface,
                full_network_config=util.get_cfg_option_bool(
                    self.ds_cfg, "apply_full_imds_network_config", True
                ),
            )
        else:
            LOG.warning("Metadata 'network' key not valid: %s.", net_md)
            return result
        self._network_config = result
        return self._network_config

    def _maybe_fetch_api_token(self, mdurls):
        """Get an API token for ECS Instance Metadata Service.

        On ECS. IMDS will always answer an API token, set
        HttpTokens=optional (default) when create instance will not forcefully
        use the security-enhanced mode (IMDSv2).

        https://api.alibabacloud.com/api/Ecs/2014-05-26/RunInstances
        """

        urls = []
        url2base = {}
        url_path = self.api_token_route
        request_method = "PUT"
        for url in mdurls:
            cur = "{0}/{1}".format(url, url_path)
            urls.append(cur)
            url2base[cur] = url

        # use the self._imds_exception_cb to check for Read errors
        LOG.debug("Fetching Ecs IMDSv2 API Token")

        response = None
        url = None
        url_params = self.get_url_params()
        try:
            url, response = uhelp.wait_for_url(
                urls=urls,
                max_wait=url_params.max_wait_seconds,
                timeout=url_params.timeout_seconds,
                status_cb=LOG.warning,
                headers_cb=self._get_headers,
                exception_cb=self._imds_exception_cb,
                request_method=request_method,
                headers_redact=self.imdsv2_token_redact,
                connect_synchronously=False,
            )
        except uhelp.UrlError:
            # We use the raised exception to interupt the retry loop.
            # Nothing else to do here.
            pass

        if url and response:
            self._api_token = response
            return url2base[url]

        # If we get here, then wait_for_url timed out, waiting for IMDS
        # or the IMDS HTTP endpoint is disabled
        return None

    def wait_for_metadata_service(self):
        mcfg = self.ds_cfg
        mdurls = mcfg.get("metadata_urls", self.metadata_urls)

        # try the api token path first
        metadata_address = self._maybe_fetch_api_token(mdurls)

        if metadata_address:
            self.metadata_address = metadata_address
            LOG.debug("Using metadata source: '%s'", self.metadata_address)
        else:
            LOG.warning("IMDS's HTTP endpoint is probably disabled")
        return bool(metadata_address)

    def crawl_metadata(self):
        """Crawl metadata service when available.

        @returns: Dictionary of crawled metadata content containing the keys:
          meta-data, user-data, vendor-data and dynamic.
        """
        if not self.wait_for_metadata_service():
            return {}
        redact = self.imdsv2_token_redact
        crawled_metadata = {}
        exc_cb = self._refresh_stale_aliyun_token_cb
        exc_cb_ud = self._skip_or_refresh_stale_aliyun_token_cb
        skip_cb = None
        exe_cb_whole_meta = self._skip_json_path_meta_path_aliyun_cb
        try:
            crawled_metadata["user-data"] = aliyun.get_instance_data(
                self.min_metadata_version,
                self.metadata_address,
                headers_cb=self._get_headers,
                headers_redact=redact,
                exception_cb=exc_cb_ud,
                item_name="user-data",
            )
            crawled_metadata["vendor-data"] = aliyun.get_instance_data(
                self.min_metadata_version,
                self.metadata_address,
                headers_cb=self._get_headers,
                headers_redact=redact,
                exception_cb=exc_cb_ud,
                item_name="vendor-data",
            )
            try:
                result = aliyun.get_instance_meta_data(
                    self.min_metadata_version,
                    self.metadata_address,
                    headers_cb=self._get_headers,
                    headers_redact=redact,
                    exception_cb=exe_cb_whole_meta,
                )
                crawled_metadata["meta-data"] = result
            except Exception:
                util.logexc(
                    LOG,
                    "Faild read json meta-data from %s "
                    "fall back directory tree style",
                    self.metadata_address,
                )
                crawled_metadata["meta-data"] = ec2.get_instance_metadata(
                    self.min_metadata_version,
                    self.metadata_address,
                    headers_cb=self._get_headers,
                    headers_redact=redact,
                    exception_cb=exc_cb,
                    retrieval_exception_ignore_cb=skip_cb,
                )
        except Exception:
            util.logexc(
                LOG,
                "Failed reading from metadata address %s",
                self.metadata_address,
            )
            return {}
        return crawled_metadata

    def _refresh_stale_aliyun_token_cb(self, msg, exception):
        """Exception handler for Ecs to refresh token if token is stale."""
        if isinstance(exception, uhelp.UrlError) and exception.code == 401:
            # With _api_token as None, _get_headers will _refresh_api_token.
            LOG.debug("Clearing cached Ecs API token due to expiry")
            self._api_token = None
        return True  # always retry

    def _skip_retry_on_codes(self, status_codes, cause):
        """Returns False if cause.code is in status_codes."""
        return cause.code not in status_codes

    def _skip_or_refresh_stale_aliyun_token_cb(self, msg, exception):
        """Callback will not retry on SKIP_USERDATA_VENDORDATA_CODES or
        if no token is available."""
        retry = self._skip_retry_on_codes(ec2.SKIP_USERDATA_CODES, exception)
        if not retry:
            return False  # False raises exception
        return self._refresh_stale_aliyun_token_cb(msg, exception)

    def _skip_json_path_meta_path_aliyun_cb(self, msg, exception):
        """Callback will not retry of whole meta_path is not found"""
        if isinstance(exception, uhelp.UrlError) and exception.code == 404:
            LOG.warning("whole meta_path is not found, skipping")
            return False
        return self._refresh_stale_aliyun_token_cb(msg, exception)

    def _get_data(self):
        if self.cloud_name != self.dsname.lower():
            return False
        if self.perform_dhcp_setup:  # Setup networking in init-local stage.
            if util.is_FreeBSD():
                LOG.debug("FreeBSD doesn't support running dhclient with -sf")
                return False
            try:
                with EphemeralIPNetwork(
                    self.distro,
                    self.distro.fallback_interface,
                    ipv4=True,
                    ipv6=False,
                ) as netw:
                    self._crawled_metadata = self.crawl_metadata()
                    LOG.debug(
                        "Crawled metadata service%s",
                        f" {netw.state_msg}" if netw.state_msg else "",
                    )

            except NoDHCPLeaseError:
                return False
        else:
            self._crawled_metadata = self.crawl_metadata()
        if not self._crawled_metadata or not isinstance(
            self._crawled_metadata, dict
        ):
            return False
        self.metadata = self._crawled_metadata.get("meta-data", {})
        self.userdata_raw = self._crawled_metadata.get("user-data", {})
        self.vendordata_raw = self._crawled_metadata.get("vendor-data", {})
        return True

    def _refresh_api_token(self, seconds=None):
        """Request new metadata API token.
        @param seconds: The lifetime of the token in seconds

        @return: The API token or None if unavailable.
        """

        if seconds is None:
            seconds = self.imdsv2_token_ttl_seconds

        LOG.debug("Refreshing Ecs metadata API token")
        request_header = {self.imdsv2_token_req_header: seconds}
        token_url = "{}/{}".format(self.metadata_address, self.api_token_route)
        try:
            response = uhelp.readurl(
                token_url,
                headers=request_header,
                headers_redact=self.imdsv2_token_redact,
                request_method="PUT",
            )
        except uhelp.UrlError as e:
            LOG.warning(
                "Unable to get API token: %s raised exception %s", token_url, e
            )
            return None
        return response.contents

    def _get_headers(self, url=""):
        """Return a dict of headers for accessing a url.

        If _api_token is unset on AWS, attempt to refresh the token via a PUT
        and then return the updated token header.
        """

        request_token_header = {
            self.imdsv2_token_req_header: self.imdsv2_token_ttl_seconds
        }
        if self.api_token_route in url:
            return request_token_header
        if not self._api_token:
            # If we don't yet have an API token, get one via a PUT against
            # api_token_route. This _api_token may get unset by a 403 due
            # to an invalid or expired token
            self._api_token = self._refresh_api_token()
            if not self._api_token:
                return {}
        return {self.imdsv2_token_put_header: self._api_token}

    def _imds_exception_cb(self, exception=None):
        """Fail quickly on proper AWS if IMDSv2 rejects API token request

        Guidance from Amazon is that if IMDSv2 had disabled token requests
        by returning a 403, or cloud-init malformed requests resulting in
        other 40X errors, we want the datasource detection to fail quickly
        without retries as those symptoms will likely not be resolved by
        retries.

        Exceptions such as requests.ConnectionError due to IMDS being
        temporarily unroutable or unavailable will still retry due to the
        callsite wait_for_url.
        """
        if isinstance(exception, uhelp.UrlError):
            # requests.ConnectionError will have exception.code == None
            if exception.code and exception.code >= 400:
                if exception.code == 403:
                    LOG.warning(
                        "Ecs IMDS endpoint returned a 403 error. "
                        "HTTP endpoint is disabled. Aborting."
                    )
                else:
                    LOG.warning(
                        "Fatal error while requesting Ecs IMDSv2 API tokens"
                    )
                raise exception
        return True


def _is_aliyun():
    return dmi.read_dmi_data("system-product-name") == ALIYUN_PRODUCT


def parse_public_keys(public_keys):
    keys = []
    for _key_id, key_body in public_keys.items():
        if isinstance(key_body, str):
            keys.append(key_body.strip())
        elif isinstance(key_body, list):
            keys.extend(key_body)
        elif isinstance(key_body, dict):
            key = key_body.get("openssh-key", [])
            if isinstance(key, str):
                keys.append(key.strip())
            elif isinstance(key, list):
                keys.extend(key)
    return keys


class DataSourceAliYunLocal(DataSourceAliYun):
    """Datasource run at init-local which sets up network to query metadata.

    In init-local, no network is available. This subclass sets up minimal
    networking with dhclient on a viable nic so that it can talk to the
    metadata service. If the metadata service provides network configuration
    then render the network configuration for that instance based on metadata.
    """

    perform_dhcp_setup = True


# Used to match classes to dependencies
datasources = [
    (DataSourceAliYunLocal, (sources.DEP_FILESYSTEM,)),  # Run at init-local
    (DataSourceAliYun, (sources.DEP_FILESYSTEM, sources.DEP_NETWORK)),
]


# Return a list of data sources that match this set of dependencies
def get_datasource_list(depends):
    return sources.list_from_depends(depends, datasources)



# Copyright (C) 2009-2010 Canonical Ltd.
# Copyright (C) 2012, 2013 Hewlett-Packard Development Company, L.P.
#
# Author: Scott Moser <scott.moser@canonical.com>
# Author: Juerg Haefliger <juerg.haefliger@hp.com>
#
# This file is part of cloud-init. See LICENSE file for license information.
"""Set Passwords: Set user passwords and enable/disable SSH password auth"""

import logging
import random
import re
import string
from typing import List

from cloudinit import features, lifecycle, subp, util
from cloudinit.cloud import Cloud
from cloudinit.config import Config
from cloudinit.config.schema import MetaSchema
from cloudinit.distros import ALL_DISTROS, Distro, ug_util
from cloudinit.log import log_util
from cloudinit.settings import PER_INSTANCE
from cloudinit.ssh_util import update_ssh_config

meta: MetaSchema = {
    "id": "cc_set_passwords",
    "distros": [ALL_DISTROS],
    "frequency": PER_INSTANCE,
    "activate_by_schema_keys": [],
}

LOG = logging.getLogger(__name__)


def get_users_by_type(users_list: list, pw_type: str) -> list:
    """either password or type: RANDOM is required, user is always required"""
    return (
        []
        if not users_list
        else [
            (item["name"], item.get("password", "RANDOM"))
            for item in users_list
            if item.get("type", "hash") == pw_type
        ]
    )


def _restart_ssh_daemon(distro: Distro, service: str, *extra_args: str):
    try:
        distro.manage_service("restart", service, *extra_args)
        LOG.debug("Restarted the SSH daemon.")
    except subp.ProcessExecutionError as e:
        LOG.warning(
            "'ssh_pwauth' configuration may not be applied. Cloud-init was "
            "unable to restart SSH daemon due to error: '%s'",
            e,
        )


def handle_ssh_pwauth(pw_auth, distro: Distro):
    """Apply sshd PasswordAuthentication changes.

    @param pw_auth: config setting from 'pw_auth'.
                    Best given as True, False, or "unchanged".
    @param distro: an instance of the distro class for the target distribution

    @return: None"""
    service = distro.get_option("ssh_svcname", "ssh")

    cfg_name = "PasswordAuthentication"

    if isinstance(pw_auth, str):
        lifecycle.deprecate(
            deprecated="Using a string value for the 'ssh_pwauth' key",
            deprecated_version="22.2",
            extra_message="Use a boolean value with 'ssh_pwauth'.",
        )
    if util.is_true(pw_auth):
        cfg_val = "yes"
    elif util.is_false(pw_auth):
        cfg_val = "no"
    else:
        bmsg = f"Leaving SSH config '{cfg_name}' unchanged."
        if pw_auth is None or pw_auth.lower() == "unchanged":
            LOG.debug("%s ssh_pwauth=%s", bmsg, pw_auth)
        else:
            LOG.warning("%s Unrecognized value: ssh_pwauth=%s", bmsg, pw_auth)
        return

    updated = update_ssh_config({cfg_name: cfg_val})
    if not updated:
        LOG.debug("No need to restart SSH service, %s not updated.", cfg_name)
        return

    if distro.uses_systemd():
        state = subp.subp(
            [
                "systemctl",
                "show",
                "--property",
                "ActiveState",
                "--value",
                service,
            ]
        ).stdout.strip()
        if state.lower() in ["active", "activating", "reloading"]:
            # This module runs Before=sshd.service. What that means is that
            # the code can only get to this point if a user manually starts the
            # network stage. While this isn't a well-supported use-case, this
            # does cause a deadlock if started via systemd directly:
            # "systemctl start cloud-init.service". Prevent users from causing
            # this deadlock by forcing systemd to ignore dependencies when
            # restarting. Note that this deadlock is not possible in newer
            # versions of cloud-init, since starting the second service doesn't
            # run the second stage in 24.3+. This code therefore exists solely
            # for backwards compatibility so that users who think that they
            # need to manually start cloud-init (why?) with systemd (again,
            # why?) can do so.
            _restart_ssh_daemon(
                distro, service, "--job-mode=ignore-dependencies"
            )
    else:
        _restart_ssh_daemon(distro, service)


def handle(name: str, cfg: Config, cloud: Cloud, args: list) -> None:
    distro: Distro = cloud.distro
    if args:
        # if run from command line, and give args, wipe the chpasswd['list']
        password = args[0]
        if "chpasswd" in cfg and "list" in cfg["chpasswd"]:
            del cfg["chpasswd"]["list"]
    else:
        password = util.get_cfg_option_str(cfg, "password", None)

    expire = True
    plist: List = []
    users_list: List = []

    if "chpasswd" in cfg:
        chfg = cfg["chpasswd"]
        users_list = util.get_cfg_option_list(chfg, "users", default=[])
        if "list" in chfg and chfg["list"]:
            lifecycle.deprecate(
                deprecated="Config key 'lists'",
                deprecated_version="22.3",
                extra_message="Use 'users' instead.",
            )
            if isinstance(chfg["list"], list):
                LOG.debug("Handling input for chpasswd as list.")
                plist = util.get_cfg_option_list(chfg, "list", plist)
            else:
                lifecycle.deprecate(
                    deprecated="The chpasswd multiline string",
                    deprecated_version="22.2",
                    extra_message="Use string type instead.",
                )
                LOG.debug("Handling input for chpasswd as multiline string.")
                multiline = util.get_cfg_option_str(chfg, "list")
                if multiline:
                    plist = multiline.splitlines()

        expire = util.get_cfg_option_bool(chfg, "expire", expire)

    if not (users_list or plist) and password:
        (users, _groups) = ug_util.normalize_users_groups(cfg, distro)
        (user, _user_config) = ug_util.extract_default(users)
        if user:
            plist = ["%s:%s" % (user, password)]
        else:
            LOG.warning("No default or defined user to change password for.")

    errors = []
    if plist or users_list:
        # This section is for parsing the data that arrives in the form of
        #   chpasswd:
        #     users:
        plist_in = get_users_by_type(users_list, "text")
        users = [user for user, _ in plist_in]
        hashed_plist_in = get_users_by_type(users_list, "hash")
        hashed_users = [user for user, _ in hashed_plist_in]
        randlist = []
        for user, _ in get_users_by_type(users_list, "RANDOM"):
            password = rand_user_password()
            users.append(user)
            plist_in.append((user, password))
            randlist.append(f"{user}:{password}")

        # This for loop is for parsing the data that arrives in the deprecated
        # form of
        #   chpasswd:
        #     list:
        # N.B. This regex is included in the documentation (i.e. the schema
        # docstring), so any changes to it should be reflected there.
        prog = re.compile(r"\$(1|2a|2y|5|6)(\$.+){2}")
        for line in plist:
            u, p = line.split(":", 1)
            if prog.match(p) is not None and ":" not in p:
                hashed_plist_in.append((u, p))
                hashed_users.append(u)
            else:
                # in this else branch, we potentially change the password
                # hence, a deviation from .append(line)
                if p == "R" or p == "RANDOM":
                    p = rand_user_password()
                    randlist.append("%s:%s" % (u, p))
                plist_in.append((u, p))
                users.append(u)
        if users:
            try:
                LOG.debug("Changing password for %s:", users)
                distro.chpasswd(plist_in, hashed=False)
            except Exception as e:
                errors.append(e)
                util.logexc(
                    LOG, "Failed to set passwords with chpasswd for %s", users
                )

        if hashed_users:
            try:
                LOG.debug("Setting hashed password for %s:", hashed_users)
                distro.chpasswd(hashed_plist_in, hashed=True)
            except Exception as e:
                errors.append(e)
                util.logexc(
                    LOG,
                    "Failed to set hashed passwords with chpasswd for %s",
                    hashed_users,
                )

        if len(randlist):
            blurb = (
                "Set the following 'random' passwords\n",
                "\n".join(randlist),
            )
            log_util.multi_log(
                "%s\n%s\n" % blurb, stderr=False, fallback_to_stdout=False
            )

        if expire:
            users_to_expire = users
            if features.EXPIRE_APPLIES_TO_HASHED_USERS:
                users_to_expire += hashed_users
            expired_users = []
            for u in users_to_expire:
                try:
                    distro.expire_passwd(u)
                    expired_users.append(u)
                except Exception as e:
                    errors.append(e)
                    util.logexc(LOG, "Failed to set 'expire' for %s", u)
            if expired_users:
                LOG.debug("Expired passwords for: %s users", expired_users)

    handle_ssh_pwauth(cfg.get("ssh_pwauth"), distro)

    if len(errors):
        LOG.debug("%s errors occurred, re-raising the last one", len(errors))
        raise errors[-1]


def rand_user_password(pwlen=20):
    if pwlen < 4:
        raise ValueError("Password length must be at least 4 characters.")

    # There are often restrictions on the minimum number of character
    # classes required in a password, so ensure we at least one character
    # from each class.
    res_rand_list = [
        random.choice(string.digits),
        random.choice(string.ascii_lowercase),
        random.choice(string.ascii_uppercase),
        random.choice(string.punctuation),
    ]

    res_rand_list.extend(
        list(
            util.rand_str(
                pwlen - len(res_rand_list),
                select_from=string.digits
                + string.ascii_lowercase
                + string.ascii_uppercase
                + string.punctuation,
            )
        )
    )
    random.shuffle(res_rand_list)
    return "".join(res_rand_list)



"""Tests here shouldn't require any sort of user data or instance setup."""

import pytest

from tests.integration_tests import releases
from tests.integration_tests.instances import IntegrationInstance
from tests.integration_tests.util import (
    get_inactive_modules,
    verify_clean_boot,
    verify_clean_log,
)


@pytest.mark.ci
class TestDefaults:
    def test_clean_log(self, class_client: IntegrationInstance):
        log = class_client.read_from_file("/var/log/cloud-init.log")
        verify_clean_log(log, ignore_deprecations=False)
        verify_clean_boot(class_client)

    def test_inactive_modules(self, class_client: IntegrationInstance):
        """Verify no errors, no deprecations and correct inactive modules in
        log.
        """
        log = class_client.read_from_file("/var/log/cloud-init.log")

        expected_inactive = {
            "apt_pipelining",
            "ansible",
            "bootcmd",
            "ca_certs",
            "chef",
            "disable_ec2_metadata",
            "disk_setup",
            "fan",
            "keyboard",
            "landscape",
            "lxd",
            "mcollective",
            "ntp",
            "package_update_upgrade_install",
            "phone_home",
            "power_state_change",
            "puppet",
            "rsyslog",
            "runcmd",
            "salt_minion",
            "snap",
            "timezone",
            "ubuntu_autoinstall",
            "ubuntu_pro",
            "ubuntu_drivers",
            "update_etc_hosts",
            "wireguard",
            "write_files",
            "write_files_deferred",
        }
        if releases.CURRENT_RELEASE >= releases.PLUCKY:
            expected_inactive.add("grub_dpkg")

        # Remove modules that run independent from user-data
        if class_client.settings.PLATFORM == "azure":
            expected_inactive.discard("disk_setup")
        elif class_client.settings.PLATFORM == "gce":
            expected_inactive.discard("ntp")
        elif class_client.settings.PLATFORM == "lxd_vm":
            if class_client.settings.OS_IMAGE == "bionic":
                expected_inactive.discard("write_files")
                expected_inactive.discard("write_files_deferred")
        elif class_client.settings.PLATFORM == "oci":
            expected_inactive.discard("update_etc_hosts")

        diff = expected_inactive.symmetric_difference(
            get_inactive_modules(log)
        )
        assert (
            not diff
        ), f"Expected inactive modules do not match, diff: {diff}"

    def test_var_log_cloud_init_output_not_world_readable(
        self, class_client: IntegrationInstance
    ):
        """
        The log can contain sensitive data, it shouldn't be world-readable.

        LP: #1918303
        """
        client = class_client
        # Check the file exists
        assert client.execute("test -f /var/log/cloud-init-output.log").ok

        # Check its permissions are as we expect
        perms, user, group = client.execute(
            "stat -c %a:%U:%G /var/log/cloud-init-output.log"
        ).split(":")
        assert "640" == perms
        assert "root" == user
        assert "adm" == group



import pytest

from tests.integration_tests import releases, util
from tests.integration_tests.instances import IntegrationInstance


def setup_custom_datasource(client: IntegrationInstance, datasource_name: str):
    client.write_to_file(
        "/etc/cloud/cloud.cfg.d/99-imds.cfg",
        f"datasource_list: [ {datasource_name}, None ]\n"
        "datasource_pkg_list: [ cisources ]",
    )
    assert client.execute("mkdir -p /usr/lib/python3/dist-packages/cisources")
    client.push_file(
        util.ASSETS_DIR / f"DataSource{datasource_name}.py",
        "/usr/lib/python3/dist-packages/cisources/"
        f"DataSource{datasource_name}.py",
    )


def verify_no_cache_boot(client: IntegrationInstance):
    log = client.read_from_file("/var/log/cloud-init.log")
    util.verify_ordered_items_in_text(
        [
            "No local datasource found",
            "running 'init'",
            "no cache found",
            "Detected DataSource",
            "TEST _get_data called",
        ],
        text=log,
    )
    util.verify_clean_boot(client)


@pytest.mark.skipif(
    not releases.IS_UBUNTU,
    reason="hardcoded dist-packages directory",
)
def test_no_cache_network_only(client: IntegrationInstance):
    """Test cache removal per boot. GH-5486

    This tests the CloudStack password reset use case. The expectation is:
    - Metadata is fetched in network timeframe only
    - Because `check_instance_id` is not defined, no cached datasource
      is found in the init-local phase, but the cache is used in the
      remaining phases due to existance of /run/cloud-init/.instance-id
    - Because `check_if_fallback_is_allowed` is not defined, cloud-init
      does NOT fall back to the pickled datasource, and will
      instead delete the cache during the init-local phase
    - Metadata is therefore fetched every boot in the network phase
    """
    setup_custom_datasource(client, "NoCacheNetworkOnly")

    # Run cloud-init as if first boot
    assert client.execute("cloud-init clean --logs")
    client.restart()

    verify_no_cache_boot(client)

    # Clear the log without clean and run cloud-init for subsequent boot
    assert client.execute("echo '' > /var/log/cloud-init.log")
    client.restart()

    verify_no_cache_boot(client)


@pytest.mark.skipif(
    not releases.IS_UBUNTU,
    reason="hardcoded dist-packages directory",
)
def test_no_cache_with_fallback(client: IntegrationInstance):
    """Test we use fallback when defined and no cache available."""
    setup_custom_datasource(client, "NoCacheWithFallback")

    # Run cloud-init as if first boot
    assert client.execute("cloud-init clean --logs")
    # Used by custom datasource
    client.execute("touch /ci-test-firstboot")
    client.restart()

    log = client.read_from_file("/var/log/cloud-init.log")
    util.verify_ordered_items_in_text(
        [
            "no cache found",
            "Detected DataSource",
            "TEST _get_data called",
            "running 'init'",
            "restored from cache with run check",
            "running 'modules:config'",
        ],
        text=log,
    )
    util.verify_clean_boot(client)

    # Clear the log without clean and run cloud-init for subsequent boot
    assert client.execute("echo '' > /var/log/cloud-init.log")
    client.execute("rm /ci-test-firstboot")
    client.restart()

    log = client.read_from_file("/var/log/cloud-init.log")
    util.verify_ordered_items_in_text(
        [
            "cache invalid in datasource",
            "Detected DataSource",
            "Restored fallback datasource from checked cache",
            "running 'init'",
            "restored from cache with run check",
            "running 'modules:config'",
        ],
        text=log,
    )
    util.verify_clean_boot(client)
